import json
import logging
from collections.abc import Callable
from typing import NamedTuple

from model_lib import Entity
from pydantic import Field, model_validator

from atlas_init.cli_tf.debug_logs import SDKRoundtrip

logger = logging.getLogger(__name__)


class StatusText(Entity):
    response_index: int
    status: int
    text: str
    duplicate_responses: int | None = None

    @property
    def id(self):
        return f"{self.status}_{self.text}"


class RequestInfo(Entity):
    path: str
    method: str
    version: str
    text: str
    responses: list[StatusText] = Field(default_factory=list)

    @property
    def id(self):
        return "__".join(part for part in (self.method, self.path, self.version, self.text) if part)


class StepRequests(Entity):
    diff_requests: list[RequestInfo] = Field(default_factory=list)
    request_responses: list[RequestInfo] = Field(default_factory=list)

    @property
    def all_requests(self):
        return self.diff_requests + self.request_responses

    def existing_request(self, info: RequestInfo) -> RequestInfo | None:
        return next((r for r in self.request_responses if r.id == info.id), None)

    def add_request(
        self,
        path: str,
        method: str,
        version: str,
        status: int,
        text: str,
        text_response: str,
        is_diff: bool,
        response_index: int,
    ):
        status_text = StatusText(status=status, text=text_response, response_index=response_index)
        info = RequestInfo(
            path=path,
            method=method,
            version=version,
            text=text,
            responses=[status_text],
        )
        if is_diff:
            self.diff_requests.append(info)
        if existing := self.existing_request(info):
            existing.responses.append(status_text)
        else:
            self.request_responses.append(info)


class RTModifier(Entity):
    version: str
    method: str
    path: str
    modification: Callable[[SDKRoundtrip], None]

    def match(self, rt: SDKRoundtrip, normalized_path: str) -> bool:
        return rt.request.method == self.method and normalized_path == self.path and rt.version == self.version


class MockRequestData(Entity):
    step_count: int
    steps: list[StepRequests] = Field(default_factory=list, init=False)
    variables: dict[str, str] = Field(default_factory=dict)

    @model_validator(mode="after")
    def set_steps(self):
        self.steps = [StepRequests() for _ in range(self.step_count)]
        return self

    def add_roundtrip(
        self,
        rt: SDKRoundtrip,
        normalized_path: str,
        normalized_text: str,
        normalized_response_text: str,
        is_diff: bool,
    ):
        step = self.steps[rt.step_number - 1]
        step.add_request(
            normalized_path,
            rt.request.method,
            rt.version,
            rt.response.status,
            normalized_text,
            normalized_response_text,
            is_diff,
            rt.resp_index,
        )

    def update_variables(self, variables: dict[str, str]) -> None:
        if missing_value := sorted(name for name, value in variables.items() if not value):
            err_msg = f"Missing values for variables: {missing_value}"
            raise ValueError(err_msg)
        changes: list[VariableChange] = []
        for name, value in variables.items():
            old_value = self.variables.get(name)
            if old_value and old_value != value:
                for suffix in range(2, 10):
                    new_name = f"{name}{suffix}"
                    old_value2 = self.variables.get(new_name, "")
                    if old_value2 and old_value2 != value:
                        continue
                    if not old_value2:
                        logger.warning(f"Adding variable {name} to {new_name}={value}")
                    change = VariableChange(name, new_name, old_value, value)
                    changes.append(change)
                    self.variables[new_name] = value
                    break
                else:
                    raise ValueError(f"Too many variables with the same name and different values: {name}")
            else:
                self.variables[name] = value
        if changes:
            raise VariablesChangedError(changes)

    def replace_text_variables(self):
        for step in self.steps:
            for request in step.all_requests:
                request.text = normalize_text(request.text, self.variables, expect_json=True)
                for response in request.responses:
                    response.text = normalize_text(response.text, self.variables, expect_json=True)

    def prune_duplicate_responses(self):
        for step in self.steps:
            for request in step.request_responses:
                pruned_responses = []
                seen_response_ids: dict[str, StatusText] = {}
                before_len = len(request.responses)
                for response in request.responses:
                    if existing_response := seen_response_ids.get(response.id):
                        if existing_response.duplicate_responses is None:
                            existing_response.duplicate_responses = 0
                        existing_response.duplicate_responses += 1
                        continue
                    seen_response_ids[response.id] = response
                    pruned_responses.append(response)
                request.responses = pruned_responses
                after_len = len(request.responses)
                duplicate_responses = before_len - after_len
                if duplicate_responses > 0:
                    logger.info(f"Pruned {duplicate_responses} duplicate responses from {request.id}")


class ApiSpecPath(Entity):
    path: str

    def variables(self, path: str) -> dict[str, str]:
        return {
            var[1:-1]: default
            for var, default in zip(self.path.split("/"), path.split("/"), strict=False)
            if var.startswith("{") and var.endswith("}")
        }

    def match(self, path: str) -> bool:
        parts_expected = self.path.split("/")
        parts_actual = path.split("/")
        if len(parts_expected) != len(parts_actual):
            return False
        for expected, actual in zip(parts_expected, parts_actual, strict=False):
            if expected == actual:
                continue
            if expected.startswith("{") and expected.endswith("}"):
                continue
            return False
        return True


def find_normalized_path(path: str, api_spec_paths: list[ApiSpecPath]) -> ApiSpecPath:
    if "?" in path:
        path = path.split("?")[0]
    path = path.rstrip("/")  # remove trailing slash
    for api_spec_path in api_spec_paths:
        if api_spec_path.match(path):
            return api_spec_path
    raise ValueError(f"Could not find path: {path}")


def normalize_text(text: str, variables: dict[str, str], *, expect_json: bool = False) -> str:
    for var, value in variables.items():
        text = text.replace(value, f"{{{var}}}")
    if not text or not expect_json:
        return text
    try:
        parsed_text = json.loads(text)
        return json.dumps(parsed_text, indent=1, sort_keys=True)
    except json.JSONDecodeError:
        logger.warning(f"Could not parse text: {text}")
        return text


def default_is_diff(rt: SDKRoundtrip) -> bool:
    return rt.request.method not in {"GET"}


class VariableChange(NamedTuple):
    var_name: str
    new_var_name: str
    old: str
    new: str


class VariablesChangedError(Exception):
    def __init__(self, changes: list[VariableChange]) -> None:
        super().__init__(f"Variables changed: {changes}")
        self.changes = changes


def create_mock_data(
    roundtrips: list[SDKRoundtrip],
    api_spec_paths: dict[str, list[ApiSpecPath]],
    is_diff: Callable[[SDKRoundtrip], bool] | None = None,
    modifiers: list[RTModifier] | None = None,
    *,
    prune_duplicates: bool = True,
) -> MockRequestData:
    steps = max(rt.step_number for rt in roundtrips)
    mock_data = MockRequestData(step_count=steps)
    is_diff = is_diff or default_is_diff
    modifiers = modifiers or []
    for rt in roundtrips:
        request_path = rt.request.path
        method = rt.request.method
        spec_path = find_normalized_path(request_path, api_spec_paths[method])
        normalized_path, normalized_text, normalized_response_text = normalize_rt(modifiers, mock_data, rt, spec_path)
        mock_data.add_roundtrip(rt, normalized_path, normalized_text, normalized_response_text, is_diff(rt))
    mock_data.replace_text_variables()
    if prune_duplicates:
        mock_data.prune_duplicate_responses()  # better to keep duplicates to stay KISS
    return mock_data


def normalize_rt(
    modifiers: list[RTModifier],
    mock_data: MockRequestData,
    rt: SDKRoundtrip,
    spec_path: ApiSpecPath,
):
    request_path = rt.request.path
    rt_variables = spec_path.variables(request_path)
    try:
        mock_data.update_variables(rt_variables)
    except VariablesChangedError as e:
        for change in e.changes:
            rt_variables.pop(change.var_name)
            rt_variables[change.new_var_name] = change.new
    normalized_path = normalize_text(request_path, rt_variables)
    for modifier in modifiers:
        if modifier.match(rt, normalized_path):
            modifier.modification(rt)
    normalized_text = normalize_text(rt.request.text, mock_data.variables, expect_json=True)
    normalized_response_text = normalize_text(rt.response.text, mock_data.variables, expect_json=True)
    return normalized_path, normalized_text, normalized_response_text



"""Inspired by: https://github.com/tmbo/questionary/blob/master/tests/utils.py"""

from dataclasses import dataclass
from typing import Callable, TypeVar

from prompt_toolkit.input.defaults import create_pipe_input
from prompt_toolkit.output import DummyOutput
from questionary import Question, checkbox
from questionary import confirm as _confirm
from questionary import select as _select
from questionary import text as _text

T = TypeVar("T")
TypedAsk = Callable[[Question, type[T]], T]

_question_asker: TypedAsk = lambda q, _: q.ask()  # noqa: E731


def confirm(prompt_text: str, *, default: bool | None = None) -> bool:
    if default is None:
        return _question_asker(_confirm(prompt_text), bool)
    return _question_asker(_confirm(prompt_text, default=default), bool)


def select_list_multiple(
    prompt_text: str,
    choices: list[str],
    default: list[str] | None = None,
) -> list[str]:
    assert choices, "choices must not be empty"
    default = default or []
    return _question_asker(checkbox(prompt_text, choices=choices), list[str]) or default


def text(
    prompt_text: str,
    default: str = "",
) -> str:
    return _question_asker(_text(prompt_text, default=default), str)


T = TypeVar("T")


def select_dict(
    prompt_text: str,
    choices: dict[str, T],
    default: str | None = None,
) -> T:
    assert choices, "choices must not be empty"
    selection = _question_asker(_select(prompt_text, default=default, choices=list(choices)), str)
    return choices[selection]


StrT = TypeVar("StrT", bound=str)


def select_list(
    prompt_text: str,
    choices: list[StrT],
    default: StrT | None = None,
) -> StrT:
    assert choices, "choices must not be empty"
    return _question_asker(_select(prompt_text, default=default, choices=choices), str)


class KeyInput:
    DOWN = "\x1b[B"
    UP = "\x1b[A"
    LEFT = "\x1b[D"
    RIGHT = "\x1b[C"
    ENTER = "\r"
    ESCAPE = "\x1b"
    CONTROLC = "\x03"
    CONTROLN = "\x0e"
    CONTROLP = "\x10"
    BACK = "\x7f"
    SPACE = " "
    TAB = "\x09"
    ONE = "1"
    TWO = "2"
    THREE = "3"


@dataclass
class question_patcher:
    responses: list[str]
    next_response: int = 0

    def __enter__(self):
        global _question_asker
        self._old_patcher = _question_asker
        _question_asker = self.ask_question
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        global _question_patcher
        _question_patcher = self._old_patcher

    def ask_question(self, q: Question, response_type: type[T]) -> T:
        q.application.output = DummyOutput()

        def run(inp) -> T:
            try:
                input_response = self.responses[self.next_response]
            except IndexError:
                raise ValueError(
                    f"Not enough responses provided. Expected {len(self.responses)}, got {self.next_response + 1} questions."
                )
            self.next_response += 1
            inp.send_text(input_response + KeyInput.ENTER + "\r")
            q.application.output = DummyOutput()
            q.application.input = inp
            return q.ask()

        with create_pipe_input() as inp:
            return run(inp)


if __name__ == "__main__":
    print(select_list("Select an option:", ["Option 1", "Option 2", "Option 3"]))  # noqa: T201
    print(  # noqa: T201
        select_dict(
            "Select an option:",
            {"Option 1": "1", "Option 2": "2", "Option 3": "3"},
            default="Option 3",
        )
    )
    print(confirm("Can you confirm?", default=True))  # noqa: T201
    print(confirm("Can you confirm?", default=False))  # noqa: T201
    print(  # noqa: T201
        select_list_multiple("Select options:", ["Option 1", "Option 2", "Option 3"], ["Option 1"])
    )
    print(text("Enter your name:", default="John Doe"))  # noqa: T201



import json
import logging
import re
from contextlib import suppress
from functools import total_ordering
from typing import Any, NamedTuple, Self

from model_lib import Entity
from pydantic import ValidationError, model_validator

logger = logging.getLogger(__name__)


def parsed(payload: str) -> tuple[dict[str, Any] | None, list | None, bool | None]:
    with suppress(ValueError):
        resp = json.loads(payload)
        if isinstance(resp, dict):
            return resp, None, None
        if isinstance(resp, list):
            return None, resp, None
    if payload.strip() in {"true", "false"}:
        return None, None, payload.strip() == "true"
    raise ValueError(f"Could not parse payload: {payload}")


class PathHeadersPayload(Entity):
    method: str
    path: str
    http_protocol: str
    headers: dict[str, str]
    text: str

    @property
    def expect_list_response(self) -> bool:
        final_path = self.path.split("/")[-1]
        if final_path in {"settings", "processArgs"}:
            return False
        return self.method == "GET" and self.path.endswith("s") and all(not c.isdigit() for c in final_path)

    @model_validator(mode="after")
    def normalize_path(self) -> Self:
        if "?" in self.path:
            self.path = self.path.split("?", 1)[0]
        return self


def parse_request(request_lines: list[str]) -> PathHeadersPayload:
    path_line, *header_lines_payload = request_lines
    headers_end = header_lines_payload.index("")
    header_lines = header_lines_payload[:headers_end]
    payload_lines = header_lines_payload[headers_end + 1 :]
    payload_end = payload_lines.index("")
    payload_lines = payload_lines[:payload_end]
    method, path, http_protocol = path_line.split(" ")
    return PathHeadersPayload(
        method=method,
        http_protocol=http_protocol,
        path=path,
        headers=dict(header_line.split(": ", 1) for header_line in header_lines),
        text="\n".join(payload_lines),
    )


class StatusHeadersResponse(Entity):
    http_protocol: str
    status: int
    status_text: str
    headers: dict[str, str]
    text: str


def parse_response(response_lines: list[str]) -> StatusHeadersResponse:
    http_protocol_status, *header_lines_response = response_lines
    http_protocol, status, status_text = http_protocol_status.split(" ", 2)
    headers_end = header_lines_response.index("")
    header_lines = header_lines_response[:headers_end]
    response = header_lines_response[headers_end + 1 :]
    return StatusHeadersResponse(
        http_protocol=http_protocol,
        status=status,  # type: ignore
        status_text=status_text,
        headers=dict(header_line.split(": ", 1) for header_line in header_lines),
        text="\n".join(response),
    )


# application/vnd.atlas.2024-08-05+json;charset=utf-8

_version_date_pattern = re.compile(r"(\d{4}-\d{2}-\d{2})")


def extract_version(content_type: str) -> str:
    if match := _version_date_pattern.search(content_type):
        return match.group(1)
    raise ValueError(f"Could not extract version from {content_type} header")


@total_ordering
class SDKRoundtrip(Entity):
    request: PathHeadersPayload
    response: StatusHeadersResponse
    resp_index: int
    step_number: int

    @property
    def id(self) -> str:
        return f"{self.request.method}_{self.request.path}_{self.version}"

    @property
    def version(self) -> str:
        content_type = self.response.headers.get("Content-Type", "v1")
        content_type_req = self.request.headers.get("Accept", "v1")
        with suppress(ValueError):
            return extract_version(content_type)
        with suppress(ValueError):
            return extract_version(content_type_req)
        raise ValueError(f"Could not extract version from req/resp: {content_type} or {content_type_req}")

    @property
    def java_method_match(self) -> bool:
        java_method = self.response.headers.get("X-Java-Method")
        if not java_method:
            return False
        java_method_final = java_method.split("::")[-1]
        final_req_path = self.request.path.split("/")[-1]
        return final_req_path.lower() in java_method_final.lower()

    @model_validator(mode="after")
    def ensure_match(self) -> Self:
        if self.java_method_match:
            return self
        req = self.request
        resp = self.response
        resp_payload_dict, resp_payload_list, __ = parsed(resp.text)
        resp_payload_dict = resp_payload_dict or {}
        want_list = req.expect_list_response
        if want_list and resp_payload_list is None and "results" not in resp_payload_dict:
            raise ValueError(f"Expected list response but got dict: {resp.text}")
        if not want_list and (resp_payload_list or "results" in resp_payload_dict):
            raise ValueError(f"Expected dict response but got list: {resp.text}")
        return self

    def __lt__(self, other) -> bool:
        if not isinstance(other, SDKRoundtrip):
            raise TypeError
        return self.resp_index < other.resp_index


MARKER_END = "-----------------------------------"
MARKER_REQUEST_START = "---[ REQUEST ]"
MARKER_RESPONSE_START = "---[ RESPONSE ]----"
MARKER_START_STEP = "Starting TestStep: "
MARKER_TEST = "Starting TestSteps: "


class FileRef(NamedTuple):
    request_index: int
    line_start: int
    line_end: int


_name_extract = re.compile(r"test_name=(\S+)")


def parse_test_name(logs: str) -> str:
    test_count = logs.count(MARKER_TEST)
    assert test_count == 1, f"Only one test is supported, found {test_count}"
    test_start = logs.index(MARKER_TEST)
    full_line = logs[test_start:].split("\n", maxsplit=1)[0]
    if match := _name_extract.search(full_line):
        return match.group(1)
    raise ValueError(f"Could not extract test name from {full_line}")


def parse_http_requests(logs: str) -> list[SDKRoundtrip]:
    """
    Problem: With requests that are done in parallel.
    An alternative is to use parallel 1 but it will be slow
    Methods: (rejected)
    1. Look for match from `path` to the something in the payload
    2. Use the X-Java-Method header to match the response with the path
    3. X-Envoy-Upstream-Service-Time to match it

    Method: (accepted)
    Can say that expected payload is either a list or a dict and if it ends with an identifier it is higher chance for a dict
    """
    test_name = parse_test_name(logs)
    logger.info(f"Finding http requests for test name: '{test_name}'")
    requests, responses = parse_raw_req_responses(logs)
    tf_step_starts = [i for i, line in enumerate(logs.splitlines()) if MARKER_START_STEP in line]
    used_responses: set[int] = set()
    responses_list: list[StatusHeadersResponse] = list(responses.values())
    sdk_roundtrips = []
    for ref, request in requests.items():
        roundtrip = match_request(used_responses, responses_list, ref, request, tf_step_starts)
        sdk_roundtrips.append(roundtrip)
        used_responses.add(roundtrip.resp_index)
    return sorted(sdk_roundtrips)


def find_step_number(ref: FileRef, step_starts: list[int]) -> int:
    for i, step_start in enumerate(reversed(step_starts)):
        if step_start < ref.line_start:
            return len(step_starts) - i
    logger.warning(f"Could not find step start for {ref}")
    return 0


def match_request(
    used_responses: set[int],
    responses_list: list[StatusHeadersResponse],
    ref: FileRef,
    request: PathHeadersPayload,
    step_starts: list[int],
) -> SDKRoundtrip:
    for i, response in enumerate(responses_list):
        if i in used_responses:
            continue
        with suppress(ValidationError):
            step_number = find_step_number(ref, step_starts)
            return SDKRoundtrip(
                request=request,
                response=response,
                resp_index=i,
                step_number=step_number,
            )
    remaining_responses = [resp for i, resp in enumerate(responses_list) if i not in used_responses]
    err_msg = f"Could not match request {request.path} ({ref}) with any response\n\n{request}\n\n\nThere are #{len(remaining_responses)} responses left that doesn't match\n{'-' * 80}\n{'\n'.join(r.text for r in remaining_responses)}"
    raise ValueError(err_msg)


def parse_raw_req_responses(
    logs: str,
) -> tuple[dict[FileRef, PathHeadersPayload], dict[FileRef, StatusHeadersResponse]]:
    # sourcery skip: dict-comprehension
    request_count = 0
    response_count = 0
    in_request = False
    in_response = False
    current_start = 0
    requests: dict[FileRef, list[str]] = {}
    responses: dict[FileRef, list[str]] = {}
    log_lines = logs.splitlines()
    for i, line in enumerate(log_lines):
        if line.startswith(MARKER_REQUEST_START):
            in_request = True
            current_start = i + 1
        elif line.startswith(MARKER_RESPONSE_START):
            in_response = True
            current_start = i + 1
        if in_request and line.startswith(MARKER_END):
            key = FileRef(request_index=request_count, line_start=current_start, line_end=i)
            requests[key] = log_lines[current_start:i]
            request_count += 1
            in_request = False
        if in_response and line.startswith(MARKER_END):
            key = FileRef(request_index=request_count, line_start=current_start, line_end=i)
            responses[key] = log_lines[current_start:i]
            response_count += 1
            in_response = False
    assert not in_request, "Request not closed"
    assert not in_response, "Response not closed"
    assert request_count == response_count, (
        f"Mismatch in request and response count: {request_count} != {response_count}"
    )
    parsed_requests = {}
    for ref, request_lines in requests.items():
        parsed_requests[ref] = parse_request(request_lines)
    parsed_responses = {}
    for ref, response_lines in responses.items():
        parsed_responses[ref] = parse_response(response_lines)
    return parsed_requests, parsed_responses



from __future__ import annotations

import logging
import re
from functools import total_ordering
from pathlib import Path
from queue import Queue
from typing import NamedTuple

from model_lib import Entity
from pydantic import Field
from zero_3rdparty.enum_utils import StrEnum

from atlas_init.cli_tf.schema_v2 import (
    SchemaAttribute,
    SchemaResource,
    SchemaV2,
    add_go_variable_names,
    as_struct_name,
    custom_object_type_name,
    extend_import_urls,
    go_fmt,
    import_lines,
    package_name,
)
from atlas_init.humps import camelize, decamelize, pascalize

logger = logging.getLogger(__name__)


class GoStdlibType(StrEnum):
    STRING = "string"
    STRING_POINTER = "*string"
    TIME = "time.Time"
    TIME_POINTER = "*time.Time"
    BOOL = "bool"
    BOOL_POINTER = "*bool"
    INT = "int"
    INT_POINTER = "*int"
    INT64 = "int64"
    INT64_POINTER = "*int64"
    FLOAT64 = "float64"
    FLOAT64_POINTER = "*float64"
    MAP_STRING = "map[string]string"
    MAP_STRING_POINTER = "*map[string]string"


_go_types = set(GoStdlibType)


def is_go_type(value: str) -> bool:
    return value in _go_types


class GoVarName(StrEnum):
    INPUT = "input"
    ITEM = "item"
    DIAGS = "diags"
    CTX = "ctx"
    ELEMENTS = "elements"
    RESP = "resp"


add_go_variable_names(list(GoVarName))


_sdk_to_tf_funcs = {
    ("*string", "string"): lambda sdk_ref: f"types.StringPointerValue({sdk_ref})",
    ("string", "string"): lambda sdk_ref: f"types.StringValue({sdk_ref})",
    (
        "*time.Time",
        "string",
    ): lambda sdk_ref: f"types.StringPointerValue(conversion.TimePtrToStringPtr({sdk_ref}))",
}
# sdk_name -> tf_name
_sdk_attribute_aliases = {
    "group_id": "project_id",
    "mongo_dbemployee_access_grant": "mongo_db_employee_access_grant",
    "mongo_dbmajor_version": "mongo_db_major_version",
    "mongo_dbversion": "mongo_db_version",
}


@total_ordering
class SDKAttribute(Entity):
    go_type: GoStdlibType | str
    json_name: str
    struct_name: str
    nested_attributes: dict[str, SDKAttribute] = Field(default_factory=dict)

    @property
    def tf_name(self) -> str:
        default = decamelize(self.json_name)
        if override := _sdk_attribute_aliases.get(default):
            return override
        return default

    @property
    def is_nested(self) -> bool:
        return not is_go_type(self.go_type) or bool(self.nested_attributes)

    @property
    def struct_type_name(self) -> str:
        assert self.is_nested
        return self.go_type.removeprefix("*").removeprefix("[]")

    def list_nested_attributes(self) -> list[SDKAttribute]:
        return [attribute for attribute in sorted(self.nested_attributes.values()) if attribute.is_nested]

    def __lt__(self, other) -> bool:
        if not isinstance(other, SDKAttribute):
            raise TypeError
        return self.json_name < other.json_name

    def as_sdk_model(self) -> SDKModel:
        assert self.is_nested
        return SDKModel(
            name=self.struct_type_name,
            attributes=self.nested_attributes,
        )


SDKAttribute.model_rebuild()


class SDKModel(Entity):
    name: str
    attributes: dict[str, SDKAttribute] = Field(default_factory=dict)
    example_files: list[str] = Field(default_factory=list)

    def lookup_tf_name(self, tf_name: str) -> SDKAttribute:
        for attribute in self.attributes.values():
            if attribute.tf_name == tf_name:
                return attribute
        raise ValueError(f"Could not find SDK attribute for {self.name}.{tf_name}")

    def list_nested_attributes(self) -> list[SDKAttribute]:
        return [attribute for attribute in sorted(self.attributes.values()) if attribute.is_nested]


json_attribute_line = re.compile(
    r"^\s+(?P<struct_name>\w+)\s+(?P<go_type>[\w\*\.\[\]]+)\s+`json:\"(?P<json_name>\w+)",
    re.M,
)

_ignored_sdk_attributes = {"href", "links"}


def parse_sdk_model(repo_path: Path, model_name: str) -> SDKModel:
    model_path = repo_path / "admin" / f"model_{decamelize(model_name)}.go"
    model = SDKModel(name=model_name)
    for match in json_attribute_line.finditer(model_path.read_text()):
        struct_name = match.group("struct_name")
        go_type = match.group("go_type")
        json_name = match.group("json_name")
        if json_name in _ignored_sdk_attributes:
            continue
        sdk_attribute = model.attributes[struct_name] = SDKAttribute(
            go_type=go_type, json_name=json_name, struct_name=struct_name
        )
        if sdk_attribute.is_nested:
            nested_model = parse_sdk_model(repo_path, sdk_attribute.struct_type_name)
            sdk_attribute.nested_attributes = nested_model.attributes
    return model


def generate_model_go(schema: SchemaV2, resource: SchemaResource, sdk_model: SDKModel) -> str:
    func_lines = sdk_to_tf_func(schema, resource, sdk_model)
    import_urls = set()
    extend_import_urls(import_urls, func_lines)
    unformatted = "\n".join(
        [
            f"package {package_name(resource.name)}",
            "",
            *import_lines(import_urls),
            "",
            *func_lines,
        ]
    )
    return go_fmt(resource.name, unformatted)


def find_schema_attribute(schema: SchemaV2, parent: SchemaResource, sdk_attribute: SDKAttribute) -> SchemaAttribute:
    err_msg = "we might need the schema to lookup sdk_attribute as an escape hatch"
    assert schema, err_msg
    return parent.lookup_tf_name(sdk_attribute.tf_name)


def find_sdk_attribute(schema_attribute: SchemaAttribute, sdk_model: SDKModel) -> SDKAttribute:
    return sdk_model.lookup_tf_name(schema_attribute.tf_name)


class SDKAndSchemaAttribute(NamedTuple):
    sdk_attribute: SDKAttribute
    schema_attribute: SchemaAttribute


def sdk_to_tf_attribute_value(
    schema_attribute: SchemaAttribute,
    sdk_attribute: SDKAttribute,
    variable_name: GoVarName = GoVarName.INPUT,
) -> str:
    key = (sdk_attribute.go_type, schema_attribute.type)
    if key in _sdk_to_tf_funcs:
        return _sdk_to_tf_funcs[key](f"{variable_name}.{sdk_attribute.struct_name}")
    raise ValueError(f"Could not find conversion function for {key}")


def sdk_to_tf_func(schema: SchemaV2, resource: SchemaResource, sdk_model: SDKModel) -> list[str]:
    struct_name = as_struct_name(resource.name, "")
    lines = [
        f"func New{struct_name}(ctx context.Context, {GoVarName.INPUT} *admin.{sdk_model.name}) (*{struct_name}, diag.Diagnostics) {{"
    ]
    nested_attributes, call_lines = call_nested_functions(schema, resource, sdk_model.list_nested_attributes())
    lines.extend(call_lines)

    lines.append(f"  return &{struct_name}{{")
    lines.extend(tf_struct_create(resource, sdk_model))
    lines.append("  }, nil")  # close return
    lines.append("}\n")  # close function

    lines.extend(process_nested_attributes(schema, nested_attributes))
    return lines


def process_nested_attributes(schema: SchemaV2, nested_attributes: Queue[SDKAndSchemaAttribute]) -> list[str]:
    lines = []
    while not nested_attributes.empty():
        sdk_attribute, schema_attribute = nested_attributes.get()
        lines.extend(sdk_to_tf_func_nested(schema, schema_attribute, sdk_attribute))
    return lines


def as_tf_struct_field_name(schema_attribute: SchemaAttribute) -> str:
    default = pascalize(schema_attribute.name)
    if default.endswith("Id"):
        return default[:-2] + "ID"
    return default


def tf_struct_create(
    resource: SchemaResource,
    sdk_model: SDKModel,
    sdk_var_name: GoVarName = GoVarName.INPUT,
) -> list[str]:
    lines = []
    for schema_attribute in resource.sorted_attributes():
        sdk_attribute = find_sdk_attribute(schema_attribute, sdk_model)
        tf_struct_field_name = as_tf_struct_field_name(schema_attribute)
        if sdk_attribute.is_nested:
            var_name = camelize(sdk_attribute.struct_name)
            lines.append(f"    {tf_struct_field_name}: {var_name},")
        else:
            lines.append(
                f"    {tf_struct_field_name}: {sdk_to_tf_attribute_value(schema_attribute, sdk_attribute, sdk_var_name)},"
            )
    return lines


def call_nested_functions(
    schema: SchemaV2, resource: SchemaResource, sdk_attributes: list[SDKAttribute]
) -> tuple[Queue[SDKAndSchemaAttribute], list[str]]:
    nested_attributes: Queue[SDKAndSchemaAttribute] = Queue()
    lines = []
    for sdk_attribute in sdk_attributes:
        schema_attribute = find_schema_attribute(schema, resource, sdk_attribute)
        var_name = camelize(sdk_attribute.struct_name)
        lines.append(
            f"  {var_name} := New{custom_object_type_name(schema_attribute)}(ctx, {GoVarName.INPUT}.{sdk_attribute.struct_name}, {GoVarName.DIAGS})"
        )
        nested_attributes.put(SDKAndSchemaAttribute(sdk_attribute, schema_attribute))
    if lines:
        lines.insert(0, "  diags := &diag.Diagnostics{}")
        lines.extend(
            [
                "  if diags.HasError() {",
                "    return nil, *diags",
                "  }",
            ]
        )
    return nested_attributes, lines


_used_refs: set[str] = set()


def sdk_to_tf_func_nested(
    schema: SchemaV2, schema_attribute: SchemaAttribute, sdk_attribute: SDKAttribute
) -> list[str]:
    if schema_attribute.schema_ref in _used_refs:
        return []
    _used_refs.add(schema_attribute.schema_ref)
    is_object = schema_attribute.type == "object"
    if is_object:
        return sdk_to_tf_func_object(schema, schema_attribute, sdk_attribute)
    return sdk_to_tf_func_list(schema, schema_attribute, sdk_attribute)


def sdk_to_tf_func_object(
    schema: SchemaV2, schema_attribute: SchemaAttribute, sdk_attribute: SDKAttribute
) -> list[str]:
    object_type_name = custom_object_type_name(schema_attribute)
    lines: list[str] = [
        f"func New{object_type_name}(ctx context.Context, {GoVarName.INPUT} *admin.{sdk_attribute.struct_type_name}, diags *diag.Diagnostics) types.Object {{",
        f"  var nilPointer *admin.{sdk_attribute.struct_type_name}",
        f"  if {GoVarName.INPUT} == nilPointer {{",
        f"    return types.ObjectNull({object_type_name}.AttrTypes)",
        "  }",
    ]
    resource = schema.ref_resource(schema_attribute.schema_ref, use_name=schema_attribute.schema_ref_name)
    nested_attributes, call_lines = call_nested_functions(schema, resource, sdk_attribute.list_nested_attributes())
    lines.extend(call_lines)
    struct_name = as_struct_name(resource.name, "")

    lines.extend(
        [
            f"  tfModel := {struct_name}{{",
            *tf_struct_create(resource, sdk_attribute.as_sdk_model()),
            "  }",
            f"  objType, diagsLocal := types.ObjectValueFrom(ctx, {object_type_name}.AttrTypes, tfModel)",
            f"  {GoVarName.DIAGS}.Append(diagsLocal...)",
            "  return objType",
            "}\n",
        ]
    )
    lines.extend(process_nested_attributes(schema, nested_attributes))
    return lines


def sdk_to_tf_func_list(schema: SchemaV2, schema_attribute: SchemaAttribute, sdk_attribute: SDKAttribute) -> list[str]:
    list_object_type = custom_object_type_name(schema_attribute)
    nested_resource = schema.ref_resource(schema_attribute.schema_ref, use_name=schema_attribute.schema_ref_name)
    lines: list[str] = [
        f"func New{list_object_type}(ctx context.Context, {GoVarName.INPUT} *[]admin.{sdk_attribute.struct_type_name}, diags *diag.Diagnostics) types.List {{",
        f"  var nilPointer *[]admin.{sdk_attribute.struct_type_name}",
        f"  if {GoVarName.INPUT} == nilPointer {{",
        f"    return types.ListNull({list_object_type})",
        "  }",
    ]
    struct_name = as_struct_name(nested_resource.name, "")
    if nested_list_attributes := [attr for attr in sdk_attribute.list_nested_attributes() if attr.is_nested]:
        logger.warning(f"Nested list attributes: {nested_list_attributes}, are not supported yet.")
    lines.extend(
        [
            f"  tfModels := make([]{struct_name}, len(*{GoVarName.INPUT}))",
            f"  for i, item := range *{GoVarName.INPUT} {{",
            f"    tfModels[i] = {struct_name}{{",
            *tf_struct_create(
                nested_resource,
                sdk_attribute.as_sdk_model(),
                sdk_var_name=GoVarName.ITEM,
            ),
            "    }",
            "  }",
            f"  listType, diagsLocal := types.ListValueFrom(ctx, {list_object_type}, tfModels)",
            f"  {GoVarName.DIAGS}.Append(diagsLocal...)",
            "  return listType",
            "}\n",
        ]
    )
    return lines



from __future__ import annotations
from typing import Any

from pydantic import BaseModel, DirectoryPath
import typer
from model_lib import parse_payload
from zero_3rdparty.iter_utils import key_equal_value_to_dict

ENV_VAR_SDK_REPO_PATH = "SDK_REPO_PATH"
option_sdk_repo_path = typer.Option(
    "", "-sdk", "--sdk-repo-path", help="the path to the sdk repo", envvar=ENV_VAR_SDK_REPO_PATH
)
option_mms_repo_path = typer.Option("", "--mms-path", help="Path to the mms directory", envvar="MMS_PATH")


class ParsedPaths(BaseModel):
    sdk_repo_path: DirectoryPath | None = None
    mms_repo_path: DirectoryPath | None = None

    @classmethod
    def from_strings(cls, *, sdk_repo_path_str: str = "", mms_path: str = "") -> ParsedPaths:
        return cls(
            sdk_repo_path=sdk_repo_path_str or None,  # type: ignore
            mms_repo_path=mms_path or None,  # type: ignore
        )


def parse_key_values(params: list[str]) -> dict[str, str]:
    return key_equal_value_to_dict(params)


def parse_key_values_any(params: list[str]) -> dict[str, Any]:
    str_dict = parse_key_values(params)
    return {k: parse_payload(v) if v.startswith(("{", "[")) else v for k, v in str_dict.items()}



import logging
import os
from concurrent.futures import ThreadPoolExecutor, wait
from enum import StrEnum
from pathlib import Path

from model_lib import Entity
from pydantic import Field

from atlas_init.cli_helper.run import run_command_is_ok_output
from atlas_init.cli_tf.go_test_run import (
    GoTestRun,
    parse_tests,
)
from atlas_init.settings.config import TestSuite
from atlas_init.settings.env_vars import AtlasInitSettings
from atlas_init.settings.path import load_dotenv

logger = logging.getLogger(__name__)


class GoTestMode(StrEnum):
    package = "package"
    individual = "individual"
    regex = "regex"


class GoEnvVars(StrEnum):
    manual = "manual"
    vscode = "vscode"


class GoTestCaptureMode(StrEnum):
    capture = "capture"
    replay = "replay"
    replay_and_update = "replay-and-update"
    no_capture = "no-capture"


def env_vars_for_capture(mode: GoTestCaptureMode) -> dict[str, str]:
    env = {}
    if mode == GoTestCaptureMode.capture:
        env["HTTP_MOCKER_CAPTURE"] = "true"
    if mode in {GoTestCaptureMode.replay, GoTestCaptureMode.replay_and_update}:
        env["HTTP_MOCKER_REPLAY"] = "true"
    if mode == GoTestCaptureMode.replay_and_update:
        env["HTTP_MOCKER_DATA_UPDATE"] = "true"
    return env


class GoTestResult(Entity):
    logs_dir: Path
    runs: dict[str, list[GoTestRun]] = Field(default_factory=dict)
    failure_names: set[str] = Field(default_factory=set)

    test_name_package_path: dict[str, Path] = Field(default_factory=dict)

    def add_test_package_path(self, test_name: str, package_path: Path):
        if old_path := self.test_name_package_path.get(test_name):
            logger.warning(f"overwriting test_name={test_name} with package_path={old_path} --> {package_path}")
        self.test_name_package_path[test_name] = package_path

    def add_test_results_all_pass(self, test_name: str, test_results: list[GoTestRun]) -> bool:
        prev_test_results = self.runs.setdefault(test_name, [])
        if prev_test_results:
            logger.warning(f"2nd time test results for {test_name}")
        for result in test_results:
            log_path = _log_path(self.logs_dir, test_name)
            result.log_path = log_path
        prev_test_results.extend(test_results)
        return all(run.is_pass for run in test_results)


def run_go_tests(
    repo_path: Path,
    settings: AtlasInitSettings,
    groups: list[TestSuite],  # type: ignore
    mode: GoTestMode | str = GoTestMode.package,
    *,
    dry_run: bool = False,
    timeout_minutes: int = 300,
    concurrent_runs: int = 20,
    re_run: bool = False,
    env_vars: GoEnvVars = GoEnvVars.vscode,
    names: set[str] | None = None,
    capture_mode: GoTestCaptureMode = GoTestCaptureMode.capture,
    use_old_schema: bool = False,
) -> GoTestResult:
    test_env = resolve_env_vars(
        settings,
        env_vars,
        capture_mode=capture_mode,
        use_old_schema=use_old_schema,
    )
    if ci_value := test_env.pop("CI", None):
        logger.warning(f"popped CI={ci_value}")
    logs_dir = settings.go_test_logs_dir
    results = GoTestResult(logs_dir=logs_dir)
    commands_to_run: dict[str, str] = {}
    for group in groups:
        if group.sequential_tests:
            logger.info(f"running individual tests sequentially as {group.name} is set to sequential_tests")
            concurrent_runs = 1
        group_commands_to_run = group_commands_for_mode(
            repo_path, mode, concurrent_runs, timeout_minutes, names, results, group
        )
        if not group_commands_to_run:
            logger.warning(f"no tests for suite: {group.name}")
            continue
        commands_to_run |= group_commands_to_run
    commands_str = "\n".join(f"'{name}': '{command}'" for name, command in sorted(commands_to_run.items()))
    logger.info(f"will run the following commands:\n{commands_str}")
    if dry_run:
        return results
    if not commands_to_run:
        logger.warning("no tests to run!")
        return results
    return _run_tests(
        results,
        repo_path,
        logs_dir,
        commands_to_run,
        test_env,
        test_timeout_s=timeout_minutes * 60,
        max_workers=concurrent_runs,
        re_run=re_run,
    )


def group_commands_for_mode(
    repo_path: Path,
    mode: GoTestMode | str,
    concurrent_runs: int,
    timeout_minutes: int,
    names: set[str] | None,
    results: GoTestResult,
    group: TestSuite,  # type: ignore
) -> dict[str, str]:
    commands_to_run: dict[str, str] = {}
    if mode == GoTestMode.package:
        name_regex = f"^({'|'.join(names)})$" if names else "^TestAcc*"
        for pkg_url in group.package_url_tests(repo_path):
            command = f"go test {pkg_url} -v -run {name_regex} -timeout {timeout_minutes}m"
            if not group.sequential_tests:
                command = f"{command} -parallel {concurrent_runs}"
            pkg_name = pkg_url.rsplit("/")[-1]
            commands_to_run[f"{group.name}-{pkg_name}"] = command
        return commands_to_run
    if mode == GoTestMode.individual:
        prefix = "TestAcc"
    else:
        logger.info(f"using {GoTestMode.regex} with {mode}")
        prefix = mode
    for pkg_url, tests in group.package_url_tests(repo_path, prefix=prefix).items():
        for name, pkg_path in tests.items():
            if names and name not in names:
                continue
            results.add_test_package_path(name, pkg_path)
            commands_to_run[name] = f"go test {pkg_url} -v -run ^{name}$ -timeout {timeout_minutes}m"
    return commands_to_run


def resolve_env_vars(
    settings: AtlasInitSettings,
    env_vars: GoEnvVars,
    *,
    capture_mode: GoTestCaptureMode,
    use_old_schema: bool,
    skip_os: bool = False,
) -> dict[str, str]:
    if env_vars == GoEnvVars.manual:
        test_env_vars = settings.manual_env_vars
    elif env_vars == GoEnvVars.vscode:
        test_env_vars = load_dotenv(settings.env_vars_vs_code)
    else:
        raise NotImplementedError(f"don't know how to load env_vars={env_vars}")
    test_env_vars |= {
        "TF_ACC": "1",
        "TF_LOG": "DEBUG",
        "MONGODB_ATLAS_PREVIEW_PROVIDER_V2_ADVANCED_CLUSTER": ("false" if use_old_schema else "true"),
    }
    test_env_vars |= env_vars_for_capture(capture_mode)
    logger.info(f"go test env-vars-extra: {sorted(test_env_vars)}")
    if not skip_os:
        test_env_vars = os.environ | test_env_vars  # os.environ on the left side, prefer explicit args
    return test_env_vars


def _run_tests(
    results: GoTestResult,
    repo_path: Path,
    logs_dir: Path,
    commands_to_run: dict[str, str],
    test_env: dict[str, str],
    test_timeout_s: int = 301 * 60,
    max_workers: int = 2,
    *,
    re_run: bool = False,
) -> GoTestResult:
    futures = {}
    actual_workers = min(max_workers, len(commands_to_run)) or 1
    with ThreadPoolExecutor(max_workers=actual_workers) as pool:
        for name, command in sorted(commands_to_run.items()):
            log_path = _log_path(logs_dir, name)
            if log_path.exists() and log_path.read_text():
                if re_run:
                    logger.info(f"moving existing logs of {name} to old dir")
                    move_logs_to_dir(logs_dir, {name}, dir_name="old")
                else:
                    logger.info(f"skipping {name} because log exists")
                    continue
            command_env = {**test_env, "TF_LOG_PATH": str(log_path)}
            future = pool.submit(
                run_command_is_ok_output,
                command=command,
                env=command_env,
                cwd=repo_path,
                logger=logger,
            )
            futures[future] = name
        done, not_done = wait(futures.keys(), timeout=test_timeout_s)
        for f in not_done:
            logger.warning(f"timeout to run command name = {futures[f]}")
    for f in done:
        name: str = futures[f]
        try:
            ok, command_out = f.result()
        except Exception:
            logger.exception(f"failed to run command for {name}")
            results.failure_names.add(name)
            continue
        try:
            parsed_tests = parse_tests(command_out.splitlines())
        except Exception:
            logger.exception(f"failed to parse tests for {name}")
            results.failure_names.add(name)
            continue
        for test in parsed_tests:
            test.log_path = _log_path(logs_dir, name)
            # todo: possible add other fields
        if not parsed_tests and not ok:
            results.failure_names.add(name)
            logger.error(f"failed to run tests for {name}: {command_out}")
            continue
        if not parsed_tests:
            logger.warning(f"failed to parse tests for {name}: {command_out}")
            continue
        if not ok:
            logger.warning(f"failing tests for {name}: {command_out}")
        if not results.add_test_results_all_pass(name, parsed_tests):
            results.failure_names.add(name)
    if failure_names := results.failure_names:
        move_logs_to_dir(logs_dir, failure_names)
        logger.error(f"failed to run tests: {sorted(failure_names)}")
    return results


def move_logs_to_dir(logs_dir: Path, names: set[str], dir_name: str = "failures"):
    new_dir = logs_dir / dir_name
    for log in logs_dir.glob("*.log"):
        if log.stem in names:
            text = log.read_text()
            assert "\n" in text
            first_line = text.split("\n", maxsplit=1)[0]
            ts = first_line.split(" ")[0]
            log.rename(new_dir / f"{ts}.{log.name}")


def _log_path(logs_dir: Path, name: str) -> Path:
    return logs_dir / f"{name}.log"



DEFAULT_EXTERNAL_SUBSTRINGS = ["aws", "azure", "google", "gcp"]
DEFAULT_INTERNAL_SUBSTRINGS = ["atlas", "mongo", "aws_region", "gcp_region", "azure_region", "cidr"]
ATLAS_PROVIDER_NAME = "mongodbatlas"



import logging
from typing import Literal

import typer
from pydantic import BaseModel
from rich.logging import RichHandler


class _LogLevel(BaseModel):
    log_level: Literal["INFO", "WARNING", "ERROR", "CRITICAL"]


def remove_secrets(message: str, secrets: list[str]) -> str:
    for secret in secrets:
        message = message.replace(secret, "***")
    return message


class SecretsHider(logging.Filter):
    def __init__(self, secrets: list[str], name: str = "") -> None:
        self.secrets = secrets
        super().__init__(name)

    def filter(self, record: logging.LogRecord) -> bool:
        record.msg = remove_secrets(record.msg, self.secrets)
        return True


dangerous_keys = ["key", "id", "secret"]
safe_keys: list[str] = ["/"]


def hide_secrets(handler: logging.Handler, secrets_dict: dict[str, str]) -> None:
    secrets_to_hide = set()
    for key, value in secrets_dict.items():
        if not isinstance(value, str):
            continue
        key_lower = key.lower()
        if key_lower in {"true", "false"} or value.lower() in {"true", "false"} or value.isdigit():
            continue
        if any(safe in key_lower for safe in safe_keys):
            continue
        if any(danger in key_lower for danger in dangerous_keys):
            secrets_to_hide.add(value)
    handler.addFilter(SecretsHider(list(secrets_to_hide), name="secrets-hider"))


def configure_logging(
    app: typer.Typer, log_level: str = "INFO", *, is_running_in_repo: bool = False
) -> logging.Handler:
    _LogLevel(log_level=log_level)  # type: ignore
    handler = RichHandler(rich_tracebacks=False, level=log_level)
    logging.basicConfig(
        level=log_level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[handler],
    )
    if not is_running_in_repo or handler.level >= logging.WARNING:
        logging.warning("using basic tracebacks/errors")
        app.pretty_exceptions_enable = False
        app.pretty_exceptions_show_locals = False

    return handler



import logging
from enum import StrEnum
from pathlib import Path
from typing import ClassVar

from model_lib import Entity, parse_payload
from pydantic import model_validator
from zero_3rdparty.dict_nested import read_nested_or_none
from zero_3rdparty.str_utils import ensure_prefix

from atlas_init.cloud.aws import REGIONS, AwsRegion
from atlas_init.humps import pascalize
from atlas_init.repos.path import current_dir

logger = logging.getLogger(__name__)


def cfn_examples_dir(repo_path: Path) -> Path:
    return repo_path / "examples"


def infer_cfn_type_name() -> str:
    cwd = current_dir()
    for json_path in cwd.glob("*.json"):
        parsed = parse_payload(json_path)
        if type_name := read_nested_or_none(parsed, "typeName"):
            assert isinstance(type_name, str)
            return type_name
    raise ValueError(f"unable to infer cfn type name in {cwd}")


class CfnType(Entity):
    MONGODB_ATLAS_CFN_TYPE_PREFIX: ClassVar[str] = "MongoDB::Atlas::"

    type_name: str
    region_filter: AwsRegion | None = None

    @classmethod
    def validate_type_region(cls, type_name: str, region: str) -> tuple[str, str | None]:
        instance = CfnType(type_name=type_name, region_filter=region or None)
        return instance.type_name, instance.region_filter

    @model_validator(mode="after")
    def ensure_type_name_prefix(self):
        self.type_name = ensure_prefix(pascalize(self.type_name), self.MONGODB_ATLAS_CFN_TYPE_PREFIX)
        return self

    @classmethod
    def resource_name(cls, type_name: str) -> str:
        return type_name.removeprefix(cls.MONGODB_ATLAS_CFN_TYPE_PREFIX).lower()


class Operation(StrEnum):
    DELETE = "delete"
    CREATE = "create"
    UPDATE = "update"


class CfnOperation(Entity):
    operaton: Operation


def cfn_type_normalized(type_name: str) -> str:
    return type_name.removeprefix(CfnType.MONGODB_ATLAS_CFN_TYPE_PREFIX).lower()


class CfnTemplateParser(Entity):
    path: Path


def validate_type_name_regions(type_name: str, region_filter: str) -> tuple[str, list[str]]:
    type_name, region_filter = CfnType.validate_type_region(type_name, region_filter)  # type: ignore
    region_filter = region_filter or ""
    if region_filter:
        logger.info(f"{type_name} in region {region_filter}")
        regions = [region_filter]
    else:
        regions = REGIONS
        logger.info(f"{type_name} in ALL regions: {regions}")
    return type_name, regions  # type: ignore



import logging
import re
from dataclasses import dataclass
from datetime import datetime
from functools import wraps
from typing import Any, AsyncIterable, Iterable, List, Optional, Type, TypeVar

from model_lib import dump_as_dict
from motor.core import AgnosticCollection
from pydantic import BaseModel
from pymongo import ASCENDING, DESCENDING, IndexModel, ReturnDocument
from pymongo.errors import DuplicateKeyError, PyMongoError
from pymongo.results import DeleteResult
from zero_3rdparty.enum_utils import StrEnum

logger = logging.getLogger(__name__)
ModelT = TypeVar("ModelT", bound=BaseModel)


class MongoUpdateOperation(StrEnum):
    """
    References:
        https://docs.mongodb.com/manual/reference/operator/update-array/
    """

    slice = "$slice"
    pop = "$pop"
    pull = "$pull"
    unset = "$unset"
    push = "$push"
    each = "$each"
    set = "$set"
    set_on_insert = "$setOnInsert"
    inc = "$inc"


UPDATE_OPERATIONS = set(MongoUpdateOperation)


def ensure_mongo_operation(updates: dict):
    """
    >>> ensure_mongo_operation({"field1": 2})
    {'$set': {'field1': 2}}
    >>> ensure_mongo_operation({MongoUpdateOperation.set: {"field1": 2}})
    {'$set': {'field1': 2}}
    >>> ensure_mongo_operation({MongoUpdateOperation.push: {"field1": 2}})
    {'$push': {'field1': 2}}
    """
    if updates.keys() - UPDATE_OPERATIONS == set():
        return updates
    return {MongoUpdateOperation.set: updates}


class MongoQueryOperation(StrEnum):
    # must be used when checking if a boolean field is false
    eq = "$eq"
    # https://stackoverflow.com/questions/18837486/query-for-boolean-field-as-not-true-e-g-either-false-or-non-existent
    ne = "$ne"
    in_ = "$in"
    # https://www.mongodb.com/docs/manual/reference/operator/query/nin/#mongodb-query-op.-nin
    nin = "$nin"
    gt = "$gt"
    gte = "$gte"
    lt = "$lt"
    lte = "$lte"
    slice = "$slice"

    @classmethod
    def boolean_or_none(cls, bool_value: bool | None) -> dict | None:
        if bool_value is None:
            return None
        return {cls.eq: True} if bool_value else {cls.ne: True}

    @classmethod
    def in_or_none(cls, options: Iterable[Any] | None) -> dict | None:
        return None if options is None else {cls.in_: list(options)}

    @classmethod
    def nin_or_none(cls, options: Iterable[Any] | None) -> dict | None:
        return None if options is None else {cls.nin: list(options)}


duplicate_key_regex = re.compile(
    r".*error collection:"
    r"\s(?P<collection_path>[-\w\d\\.]+)"
    r"\sindex:\s"
    r"(?P<index_name>[\w_\\.\d]+)"
    r"\sdup key.*?"
    r'(?P<dup_key_value>("?[\\.\w_\d]+"?)|(null))'
)


@dataclass
class MongoConstraintDetails:
    collection_path: str
    index_name: str
    dup_key_value: Optional[str]

    def __post_init__(self):
        if self.dup_key_value:
            self.dup_key_value = self.dup_key_value.strip('"')
        if self.dup_key_value == "null":
            self.dup_key_value = None


def parse_error(error: PyMongoError) -> Optional[MongoConstraintDetails]:
    """
    >>> raw = 'E11000 duplicate key error collection: dev_situation.Robot index: _id_ dup key: { : "mw_wheel_id" }'
    >>> parse_error(raw)
    MongoConstraintDetails(collection_path='dev_situation.Robot', index_name='_id_', dup_key_value='mw_wheel_id')
    ''
    """
    error_str = str(error)
    for m in duplicate_key_regex.finditer(error_str):
        constraints = MongoConstraintDetails(**m.groupdict())
        if isinstance(error, DuplicateKeyError):
            _, constraints.dup_key_value = error.details["keyValue"].popitem()  # type: ignore
        return constraints
    logger.warning(f"unknown pymongo error:{error}")


class MongoConstraintError(Exception):
    def __init__(self, details: MongoConstraintDetails):
        self.details: MongoConstraintDetails = details


T = TypeVar("T")

ConstraintSubT = TypeVar("ConstraintSubT", bound=MongoConstraintError)


def raise_mongo_constraint_error(f: T = None, *, cls: Type[ConstraintSubT] = MongoConstraintError) -> T:
    def decorator(f: T):
        @wraps(f)  # type: ignore
        async def inner(*args, **kwargs):
            try:
                return await f(*args, **kwargs)  # type: ignore
            except PyMongoError as e:
                if details := parse_error(e):
                    raise cls(details) from e
                raise e

        return inner

    return decorator(f) if f else decorator  # type: ignore


def dump_with_id(
    model: BaseModel,
    id: str = "",
    dt_keys: Optional[List[str]] = None,
    property_keys: Optional[List[str]] = None,
    exclude: Optional[set[str]] = None,
) -> dict:
    """
    Warning:
        If you want to index on datetime, you have to set them afterwards
        As they will be dumped as strings
    """
    raw = dump_as_dict(model) if exclude is None else dump_as_dict(model.model_dump(exclude=exclude))
    if id:
        raw["_id"] = id
    if dt_keys:
        for key in dt_keys:
            raw[key] = getattr(model, key)
    if property_keys:
        for key in property_keys:
            raw[key] = getattr(model, key)
    return raw


async def create_or_replace(collection: AgnosticCollection, raw: dict) -> bool:
    """
    Returns:
        is_new: bool
    """
    result = await collection.replace_one({"_id": raw["_id"]}, raw, upsert=True)
    return bool(result.upserted_id)


async def find_one_and_update(
    collection: AgnosticCollection,
    id: str,
    updates: dict,
    return_raw_after: bool = True,
    upsert: bool = False,
    **query,
) -> Optional[dict]:
    """
    Warning:
        pops the "_id" from serialize_lib
    """
    return_doc = ReturnDocument.AFTER if return_raw_after else ReturnDocument.BEFORE
    updates = ensure_mongo_operation(updates)
    raw = await collection.find_one_and_update({"_id": id, **query}, updates, return_document=return_doc, upsert=upsert)
    if raw:
        raw.pop("_id", None)
        return raw


def microsecond_compare(mongo_dt: datetime, dt: datetime) -> bool:
    """Mongo only stores milliseconds since epoch
    https://stackoverflow.com/questions/39963143/why-is-there-a-difference-
    between-the-stored-and-queried-time-in-mongo-database."""
    with_microseconds = mongo_dt.replace(microsecond=dt.microsecond)
    return with_microseconds == dt and (mongo_dt - dt).total_seconds() < 0.001


def safe_key(key: str) -> str:
    return key.replace(".", "_DOT_")


def replace_dot_keys(values: dict) -> dict:
    """avoid InvalidDocument("key 'dev.amironenko' must not contain '.'")"""
    return {safe_key(key): value for key, value in values.items()}


def decode_delete_count(result: DeleteResult) -> int:
    return result.deleted_count


def push_and_limit_length_update(field_name: str, new_value: Any, max_size: int) -> dict:
    return {
        MongoUpdateOperation.push: {
            field_name: {
                MongoUpdateOperation.each: [new_value],
                MongoUpdateOperation.slice: -max_size,
            }
        }
    }


def index_dec(column: str) -> IndexModel:
    return IndexModel([(column, DESCENDING)])


def query_and_sort(collection: AgnosticCollection, query: dict, sort_col: str, desc: bool) -> AsyncIterable[dict]:
    sort_order = DESCENDING if desc else ASCENDING
    return collection.find(query).sort(sort_col, sort_order)



from contextlib import suppress
import logging
from pathlib import Path
from ask_shell import confirm, run_and_wait
from model_lib import dump, parse_model
import typer
from zero_3rdparty.file_utils import clean_dir, copy, ensure_parents_write_text
from atlas_init.cli_args import ParsedPaths, option_sdk_repo_path, option_mms_repo_path
from atlas_init.cli_tf.openapi import OpenapiSchema

_go_mod_line = "replace go.mongodb.org/atlas-sdk/v20250312005 v20250312005.0.0 => ../atlas-sdk-go"
logger = logging.getLogger(__name__)


def go(
    mms_path_str: str = option_mms_repo_path,
    sdk_repo_path_str: str = option_sdk_repo_path,
    mms_branch: str = typer.Option("master", "--mms-branch", help="Branch to use for mms"),
    skip_mms_openapi: bool = typer.Option(
        False, "-smms", "--skip-mms-openapi", help="Skip mms openapi generation, use existing file instead"
    ),
):
    paths = ParsedPaths.from_strings(sdk_repo_path_str=sdk_repo_path_str, mms_path=mms_path_str)
    mms_path = paths.mms_repo_path
    assert mms_path, "mms_path is required"
    sdk_path = paths.sdk_repo_path
    assert sdk_path, "sdk_path is required"
    openapi_path = safe_openapi_path(mms_path) if skip_mms_openapi else generate_openapi_spec(mms_path, mms_branch)
    openapi_path = transform_openapi(openapi_path, sdk_path / "openapi/openapi-mms.yaml")
    generate_go_sdk(sdk_path, openapi_path)
    confirm(f"Have you remembered to add to your go.mod file: {_go_mod_line}")


def transform_openapi(old: Path, dest_path: Path) -> Path:
    api_spec = parse_model(old, t=OpenapiSchema)
    new_api_spec = api_spec.model_dump()
    for path in api_spec.paths.keys():
        for method_name, method in api_spec.methods_with_name(path):
            responses = method.get("responses", {})
            for code, multi_responses in responses.items():
                with suppress(AlreadySingleVersion):
                    new_api_spec["paths"][path][method_name]["responses"][code]["content"] = use_a_single_version(
                        multi_responses, api_spec, path
                    )
            if request_body := method.get("requestBody", {}):
                with suppress(AlreadySingleVersion):
                    new_api_spec["paths"][path][method_name]["requestBody"]["content"] = use_a_single_version(
                        request_body, api_spec, path
                    )
    dest_yaml = dump(new_api_spec, "yaml")
    ensure_parents_write_text(dest_path, dest_yaml)
    return dest_path


class AlreadySingleVersion(Exception):
    pass


def use_a_single_version(multi_content: dict, api_spec: OpenapiSchema, path: str) -> dict[str, dict]:
    if api_versions := api_spec._unpack_schema_versions(multi_content):
        if len(api_versions) > 1:
            latest_version = max(api_versions)
            last_header = f"application/vnd.atlas.{latest_version}+json"
            old_content = multi_content["content"]
            assert last_header in old_content, f"failed to find {last_header} for {path} in {old_content.keys()}"
            return {last_header: old_content[last_header]}
    raise AlreadySingleVersion


def generate_openapi_spec(mms_path: Path, mms_branch: str) -> Path:
    run_and_wait(f"git stash && git checkout {mms_branch}", cwd=mms_path)
    bazelisk_bin_run = run_and_wait("mise which bazelisk", cwd=mms_path)
    bazelisk_bin = bazelisk_bin_run.stdout_one_line
    assert Path(bazelisk_bin).exists(), f"not found {bazelisk_bin}"
    openapi_run = run_and_wait(f"{bazelisk_bin} run //server:mms-openapi", cwd=mms_path, print_prefix="mms-openapi")
    assert openapi_run.clean_complete, f"failed to run {openapi_run}"
    return safe_openapi_path(mms_path)


def safe_openapi_path(mms_path: Path) -> Path:
    openapi_path = mms_path / "server/openapi/services/openapi-mms.json"
    assert openapi_path.exists(), f"not found {openapi_path}"
    return openapi_path


def generate_go_sdk(repo_path: Path, openapi_path: Path) -> None:
    SDK_FOLDER = repo_path / "admin"
    clean_dir(SDK_FOLDER, recreate=True)
    generate_script = repo_path / "tools/scripts/generate.sh"
    assert generate_script.exists(), f"not found {generate_script}"
    openapi_folder = repo_path / "openapi"
    openapi_dest_path = openapi_folder / openapi_path.name
    if openapi_path != openapi_dest_path:
        copy(openapi_path, openapi_dest_path)
    generate_env = {
        "OPENAPI_FOLDER": str(openapi_folder),
        "OPENAPI_FILE_NAME": openapi_path.name,
        "SDK_FOLDER": str(SDK_FOLDER),
    }
    run_and_wait(f"{generate_script}", cwd=repo_path / "tools", env=generate_env, print_prefix="go sdk create")
    mockery_script = repo_path / "tools/scripts/generate_mocks.sh"
    run_and_wait(f"{mockery_script}", cwd=repo_path / "tools", print_prefix="go sdk mockery")



from __future__ import annotations

import logging
import re
from collections import defaultdict, deque
from collections.abc import Callable
from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum
from functools import total_ordering
from pathlib import Path
from typing import NamedTuple, TypeAlias

import humanize
from model_lib import Entity, utc_datetime, utc_datetime_ms
from pydantic import Field, model_validator
from zero_3rdparty.datetime_utils import utc_now

from atlas_init.repos.path import go_package_prefix

logger = logging.getLogger(__name__)


class GoTestStatus(StrEnum):
    RUN = "RUN"
    PAUSE = "PAUSE"
    NAME = "NAME"
    PASS = "PASS"  # noqa: S105 #nosec
    FAIL = "FAIL"
    SKIP = "SKIP"
    CONT = "CONT"
    TIMEOUT = "TIMEOUT"
    PKG_OK = "ok"

    @classmethod
    def is_running(cls, status: GoTestStatus) -> bool:
        return status in {cls.RUN, cls.PAUSE, cls.NAME, cls.CONT}

    @classmethod
    def is_running_but_not_paused(cls, status: GoTestStatus) -> bool:
        return status != cls.PAUSE and cls.is_running(status)

    @classmethod
    def is_pass_or_fail(cls, status: GoTestStatus) -> bool:
        return status in {cls.PASS, cls.FAIL, cls.TIMEOUT}  # TIMEOUT is considered a failure in this context


class GoTestContextStep(Entity):
    name: str


class GoTestContext(Entity):
    """Abstraction on WorkflowJob to also support local runs"""

    name: str
    created_at: utc_datetime = Field(default_factory=utc_now)
    steps: list[GoTestContextStep] = Field(default_factory=list)
    html_url: str = "http://localhost"

    @classmethod
    def from_local_run(cls, name: str, steps: list[GoTestContextStep]) -> GoTestContext:
        raise NotImplementedError
        # return cls(name=name, steps=steps)


def extract_group_name(log_path: Path | None) -> str:
    """
    >>> extract_group_name(
    ...     Path(
    ...         "40216340925_tests-1.11.x-latest_tests-1.11.x-latest-false_search_deployment.txt"
    ...     )
    ... )
    'search_deployment'
    >>> extract_group_name(None)
    ''
    """
    if log_path is None:
        return ""
    if "-" not in log_path.name:
        return ""
    last_part = log_path.stem.split("-")[-1]
    return "_".join(last_part.split("_")[1:]) if "_" in last_part else last_part


def parse_tests(
    log_lines: list[str],
) -> list[GoTestRun]:
    context = ParseContext()
    parser = wait_for_relvant_line
    for line in log_lines:
        parser = parser(line, context)
        context.last_lines.append(line)
    result = context.finish_parsing()
    return result.tests


class GoTestRuntimeStats(NamedTuple):
    slowest_seconds: float
    average_seconds: float | None
    name_with_package: str
    runs: list[GoTestRun]

    @property
    def slowest_duration(self) -> str:
        return humanize.naturaldelta(self.slowest_seconds)

    @property
    def average_duration(self) -> str:
        if self.average_seconds is None:
            return ""
        return humanize.naturaldelta(self.average_seconds)


class GoTestLastPassStat(NamedTuple):
    pass_ts: utc_datetime
    name_with_package: str
    pass_when: str
    last_pass: GoTestRun


@total_ordering
class GoTestRun(Entity):
    name: str
    status: GoTestStatus = GoTestStatus.RUN
    ts: utc_datetime_ms
    output_lines: list[str] = Field(default_factory=list)
    finish_ts: utc_datetime_ms | None = None
    run_seconds: float | None = Field(default=None, init=False)

    package_url: str | None = Field(default=None, init=False)

    log_path: Path | None = Field(default=None, init=False)
    env: str | None = Field(default=None, init=False)
    branch: str | None = Field(default=None, init=False)
    resources: list[str] = Field(default_factory=list, init=False)
    job_url: str | None = Field(default=None, init=False)

    def __lt__(self, other) -> bool:
        if not isinstance(other, GoTestRun):
            raise TypeError
        return (self.ts, self.name) < (other.ts, other.name)

    @property
    def id(self) -> str:
        return f"{self.ts.isoformat()}-{self.name}"

    @property
    def when(self) -> str:
        return humanize.naturaltime(self.ts)

    @property
    def runtime_human(self) -> str:
        if seconds := self.run_seconds:
            return humanize.naturaldelta(seconds)
        return "unknown"

    @property
    def output_lines_str(self) -> str:
        return "\n".join(self.output_lines)

    @property
    def is_failure(self) -> bool:
        return self.status in {GoTestStatus.FAIL, GoTestStatus.TIMEOUT}

    @property
    def is_skipped(self) -> bool:
        return self.status == GoTestStatus.SKIP

    @property
    def is_pass(self) -> bool:
        return self.status == GoTestStatus.PASS

    @property
    def group_name(self) -> str:
        return extract_group_name(self.log_path)

    def package_rel_path(self, repo_path: Path) -> str:
        if url := self.package_url:
            prefix = go_package_prefix(repo_path)
            return url.removeprefix(prefix).rstrip("/")
        return ""

    @property
    def name_with_package(self) -> str:
        if self.package_url:
            return f"{self.package_url.split('/')[-1]}/{self.name}"
        return self.name

    @property
    def full_name(self) -> str:
        if self.package_url and self.group_name:
            return f"{self.group_name}/{self.package_url.split('/')[-1]}/{self.name}"
        if self.group_name:
            return f"{self.group_name}/{self.name}"
        return self.name_with_package

    @classmethod
    def group_by_name_package(cls, tests: list[GoTestRun]) -> dict[str, list[GoTestRun]]:
        grouped = defaultdict(list)
        for test in tests:
            grouped[test.name_with_package].append(test)
        return grouped

    @classmethod
    def pass_rate_or_skip_reason(cls, tests: list[GoTestRun], *, include_single_run: bool = False) -> tuple[float, str]:
        if not tests:
            return 0.0, "No tests"
        fail_count = sum(test.is_pass for test in tests)
        total_count = sum(GoTestStatus.is_pass_or_fail(test.status) for test in tests)
        if not include_single_run and total_count == 1:
            return 0.0, "Only one test and include_single_run is False"
        if total_count == 0:
            return 0.0, "No pass or fail tests"
        return fail_count / total_count, ""

    @classmethod
    def last_pass(cls, tests: list[GoTestRun]) -> str:
        last_pass = max((test for test in tests if test.is_pass), default=None)
        return last_pass.when if last_pass else ""

    @classmethod
    def last_pass_stats(cls, tests: list[GoTestRun], *, max_tests: int = 10) -> list[GoTestLastPassStat]:
        """Returns"""
        pass_stats: list[GoTestLastPassStat] = []
        for name_with_package, name_tests in cls.group_by_name_package(tests).items():
            has_passes = bool(sum(test.is_pass for test in name_tests))
            if not has_passes:
                continue
            has_failures = bool(sum(test.is_failure for test in name_tests))
            if not has_failures:
                continue
            last_pass_test = max((test for test in name_tests if test.is_pass))
            finish_ts = last_pass_test.finish_ts
            assert finish_ts is not None, f"last_pass {last_pass_test} has no finish_ts"
            pass_stats.append(GoTestLastPassStat(finish_ts, name_with_package, last_pass_test.when, last_pass_test))
        return sorted(pass_stats)[:max_tests]

    @classmethod
    def lowest_pass_rate(
        cls, tests: list[GoTestRun], *, max_tests: int = 10, include_single_run: bool = False
    ) -> list[tuple[float, str, list[GoTestRun]]]:
        tests_with_pass_rates = []
        grouped = cls.group_by_name_package(tests)
        for name, tests in grouped.items():
            pass_rate, skip_reason = cls.pass_rate_or_skip_reason(tests, include_single_run=include_single_run)
            if skip_reason or pass_rate == 1.0:
                continue
            tests_with_pass_rates.append((pass_rate, name, tests))
        return sorted(tests_with_pass_rates)[:max_tests]

    @classmethod
    def run_delta(cls, tests: list[GoTestRun]) -> str:
        if not tests:
            return "No tests"
        run_dates = {run.ts.date() for run in tests}
        if len(run_dates) == 1:
            return f"on {run_dates.pop().strftime('%Y-%m-%d')}"
        return f"from {min(run_dates).strftime('%Y-%m-%d')} to {max(run_dates).strftime('%Y-%m-%d')}"

    @classmethod
    def slowest_tests(cls, tests: list[GoTestRun], *, max_tests: int = 10) -> list[GoTestRuntimeStats]:
        def run_time(test: GoTestRun) -> float:
            return test.run_seconds or 0.0

        slowest_tests = sorted(tests, key=run_time, reverse=True)
        stats = []
        grouped_by_name = cls.group_by_name_package(slowest_tests)
        for slow_test in slowest_tests:
            if slow_test.name_with_package not in grouped_by_name:
                continue  # already processed
            runs = grouped_by_name.pop(slow_test.name_with_package)
            slowest_seconds = max(run_time(test) for test in runs)
            if slowest_seconds < 0.1:  # ignore tests less than 0.1 seconds
                return stats
            average_seconds = sum(run_time(test) for test in runs) / len(runs) if len(runs) > 1 else None
            stats.append(
                GoTestRuntimeStats(
                    slowest_seconds=slowest_seconds,
                    average_seconds=average_seconds,
                    name_with_package=slow_test.name_with_package,
                    runs=runs,
                )
            )
            if len(stats) >= max_tests:
                return stats
        return stats


class ParseResult(Entity):
    tests: list[GoTestRun] = Field(default_factory=list)

    @model_validator(mode="after")
    def ensure_all_tests_completed(self) -> ParseResult:
        if incomplete_tests := [
            f"{test.name}-{test.status}" for test in self.tests if GoTestStatus.is_running(test.status)
        ]:
            raise ValueError(f"some tests are not completed: {incomplete_tests}")
        if no_package_tests := [(test.name, test.log_path) for test in self.tests if test.package_url is None]:
            raise ValueError(f"some tests do not have package name: {no_package_tests}")
        test_names = {test.name for test in self.tests}
        test_group_names = {name.split("/")[0] for name in test_names if "/" in name}
        self.tests = [test for test in self.tests if test.name not in test_group_names]
        return self


@dataclass
class ParseContext:
    tests: list[GoTestRun] = field(default_factory=list)

    current_output: list[str] = field(default_factory=list, init=False)
    current_test_name: str = ""  # used for debugging and breakpoints
    last_lines: deque = field(default_factory=lambda: deque(maxlen=10), init=False)

    def add_output_line(self, line: str) -> None:
        if is_blank_line(line) and self.current_output and is_blank_line(self.current_output[-1]):
            return  # avoid two blank lines in a row
        self._add_line(line)

    def _add_line(self, line: str) -> None:
        logger.debug(f"adding line to {self.current_test_name}: {line}")
        self.current_output.append(line)

    def start_test(self, test_name: str, start_line: str, ts: str) -> None:
        run = GoTestRun(name=test_name, ts=ts)  # type: ignore
        self.tests.append(run)
        self.continue_test(test_name, start_line)

    def continue_test(self, test_name: str, line: str) -> None:
        test = self.find_unfinished_test(test_name)
        self.current_output = test.output_lines
        self.current_test_name = test_name
        self._add_line(line)

    def find_unfinished_test(self, test_name: str) -> GoTestRun:
        test = next(
            (test for test in self.tests if test.name == test_name and GoTestStatus.is_running(test.status)),
            None,
        )
        assert test is not None, f"test {test_name} not found in context"
        return test

    def set_package(self, pkg_name: str) -> None:
        for test in self.tests:
            if test.package_url is None:
                test.package_url = pkg_name

    def finish_test(
        self,
        test_name: str,
        status: GoTestStatus,
        ts: str,
        end_line: str,
        run_seconds: float | None,
        extra_lines: list[str] | None = None,
    ) -> None:
        test = self.find_unfinished_test(test_name)
        test.status = status
        test.finish_ts = datetime.fromisoformat(ts)
        if extra_lines:
            test.output_lines.extend(extra_lines)
        test.output_lines.append(end_line)
        test.run_seconds = run_seconds

    def finish_parsing(self) -> ParseResult:
        return ParseResult(tests=self.tests)


LineParserT: TypeAlias = Callable[[str, ParseContext], "LineParserT"]


def ts_pattern(name: str) -> str:
    return r"(?P<%s>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+Z?)\s*" % name


blank_pattern = re.compile(ts_pattern("ts") + r"$", re.MULTILINE)


def is_blank_line(line: str) -> bool:
    return blank_pattern.match(line) is not None


_one_or_more_digit_or_star_pattern = r"(\d|\*)+"  # due to Github secrets, some digits can be replaced with "*"
runtime_pattern_no_parenthesis = (
    rf"(?P<runtime>{_one_or_more_digit_or_star_pattern}\.{_one_or_more_digit_or_star_pattern})s"
)
runtime_pattern = rf"\({runtime_pattern_no_parenthesis}\)"


ignore_line_pattern = [
    re.compile(ts_pattern("ts") + r"\s" + ts_pattern("ts2")),
    # 2025-04-29T00:44:02.9968072Z   error=
    # 2025-04-29T00:44:02.9968279Z   | exit status 1
    re.compile(ts_pattern("ts") + r"error=\s*$", re.MULTILINE),
    re.compile(ts_pattern("ts") + r"\|"),
]

status_patterns = [
    (GoTestStatus.RUN, re.compile(ts_pattern("ts") + r"=== RUN\s+(?P<name>\S+)")),
    (GoTestStatus.PAUSE, re.compile(ts_pattern("ts") + r"=== PAUSE\s+(?P<name>\S+)")),
    (GoTestStatus.NAME, re.compile(ts_pattern("ts") + r"=== NAME\s+(?P<name>\S+)")),
    (GoTestStatus.CONT, re.compile(ts_pattern("ts") + r"=== CONT\s+(?P<name>\S+)")),
    (
        GoTestStatus.PASS,
        re.compile(ts_pattern("ts") + r"--- PASS: (?P<name>\S+)\s+" + runtime_pattern),
    ),
    (
        GoTestStatus.FAIL,
        re.compile(ts_pattern("ts") + r"--- FAIL: (?P<name>\S+)\s" + runtime_pattern),
    ),
    (
        GoTestStatus.SKIP,
        re.compile(ts_pattern("ts") + r"--- SKIP: (?P<name>\S+)\s+" + runtime_pattern),
    ),
    (
        # 2025-06-06T05:30:18.9060127Z 		TestAccClusterAdvancedCluster_replicaSetAWSProvider (4h28m7s)
        GoTestStatus.TIMEOUT,
        re.compile(ts_pattern("ts") + r"\s+(?P<name>\S+)\s\((?P<hours>\d+)?h?(?P<minutes>\d+)?m(?P<seconds>\d+)?s\)"),
    ),
]
package_patterns = [
    (
        GoTestStatus.FAIL,
        re.compile(ts_pattern("ts") + r"FAIL\s+(?P<package_url>\S+)\s+" + runtime_pattern_no_parenthesis),
    ),
    (
        GoTestStatus.FAIL,
        re.compile(ts_pattern("ts") + r"FAIL\s+(?P<package_url>\S+)\s+\(cached\)"),
    ),
    (
        GoTestStatus.PKG_OK,
        re.compile(ts_pattern("ts") + r"ok\s+(?P<package_url>\S+)\s+" + runtime_pattern_no_parenthesis),
    ),
    (
        GoTestStatus.PKG_OK,
        re.compile(ts_pattern("ts") + r"ok\s+(?P<package_url>\S+)\s+\(cached\)"),
    ),
]


def line_match_status_pattern(
    line: str,
    context: ParseContext,
) -> GoTestStatus | None:
    for status, pattern in status_patterns:
        if pattern_match := pattern.match(line):
            test_name = pattern_match.group("name")
            assert test_name, f"test name not found in line: {line} when pattern matched {pattern}"
            ts = pattern_match.group("ts")
            assert ts, f"timestamp not found in line: {line} when pattern matched {pattern}"
            match status:
                case GoTestStatus.RUN:
                    context.start_test(test_name, line, ts)
                case GoTestStatus.NAME | GoTestStatus.CONT:
                    context.continue_test(test_name, line)
                case GoTestStatus.PAUSE:
                    return status  # do nothing
                case GoTestStatus.TIMEOUT:
                    hours = pattern_match.group("hours")
                    minutes = pattern_match.group("minutes")
                    seconds = pattern_match.group("seconds")
                    run_seconds = (
                        (int(hours) * 3600 if hours else 0)
                        + (int(minutes) * 60 if minutes else 0)
                        + (int(seconds) if seconds else 0)
                    )
                    last_two_lines = list(context.last_lines)[-2:]
                    context.finish_test(test_name, status, ts, line, run_seconds, extra_lines=last_two_lines)
                case GoTestStatus.PASS | GoTestStatus.FAIL | GoTestStatus.SKIP:
                    run_time = pattern_match.group("runtime")
                    assert run_time, (
                        f"runtime not found in line with status={status}: {line} when pattern matched {pattern}"
                    )
                    seconds, milliseconds = run_time.split(".")
                    if "*" in seconds:
                        run_seconds = None
                    else:
                        run_seconds = int(seconds) + int(milliseconds.replace("*", "0")) / 1000
                    context.finish_test(test_name, status, ts, line, run_seconds)
            return status
    for pkg_status, pattern in package_patterns:
        if pattern_match := pattern.match(line):
            pkg_name = pattern_match.group("package_url")
            assert pkg_name, f"package_url not found in line: {line} when pattern matched {pattern}"
            context.set_package(pkg_name)
            return pkg_status
    return None


def wait_for_relvant_line(
    line: str,
    context: ParseContext,
) -> LineParserT:
    status = line_match_status_pattern(line, context)
    if status and GoTestStatus.is_running_but_not_paused(status):
        return add_output_line
    return wait_for_relvant_line


def add_output_line(
    line: str,
    context: ParseContext,
) -> LineParserT:
    for pattern in ignore_line_pattern:
        if pattern.match(line):
            return add_output_line
    status: GoTestStatus | None = line_match_status_pattern(line, context)
    if status is None:
        context.add_output_line(line)
        return add_output_line
    if status and GoTestStatus.is_running_but_not_paused(status):
        return add_output_line
    return wait_for_relvant_line



import logging
import re
from typing import Annotated, Literal

import typer
from model_lib import Entity
from pydantic import BeforeValidator, model_validator

logger = logging.getLogger(__name__)


type ChangelogGroup = Literal["rs", "ds", "dsp"]
CHANGELOG_GROUPS = {"rs", "ds", "dsp"}
_group_prefixes = {
    "rs": "resource/",
    "ds": "data-source/",
    "dsp": "data-source/",
}
_group_suffixes = {
    "rs": "",
    "ds": "",
    "dsp": "s",
}
type ChangelogType = Literal["feat", "fix"]
CHANGELOG_SEPARATOR = "\n\n"
_types_headers = {"feat": "enhancement", "fix": "bug"}
_line_regex = re.compile(r"(?P<type>feat|fix)\((?P<groups>[^\]]+)\)\:\s(?P<text>.*)")


def as_group_name(resource_name: str, group: ChangelogGroup):
    prefix = _group_prefixes[group]
    suffix = _group_suffixes[group]
    return f"{prefix}{resource_name}{suffix}"


def changelog_header(group: ChangelogType) -> str:
    return f"release-note:{_types_headers[group]}"


def split_on_plus(groups: str | list[str]) -> list[str]:
    return groups.split("+") if isinstance(groups, str) else groups


type GroupsValidation = Annotated[str, BeforeValidator(split_on_plus)]


class ChangelogPart(Entity):
    type: ChangelogType
    groups: str
    text: str
    resource_name: str

    @property
    def parsed_groups(self) -> list[ChangelogGroup]:
        groups = set(self.groups.split("+"))
        if invalid_groups := groups - CHANGELOG_GROUPS:
            raise ValueError(f"found invalid groups {invalid_groups}, only {CHANGELOG_GROUPS} are valid")
        return sorted(groups)  # type: ignore

    @model_validator(mode="after")
    def ensure_parsed_groups(self):
        assert self.parsed_groups
        return self

    def as_changelog(self) -> str:
        parts = []
        for group in self.parsed_groups:
            header = changelog_header(self.type)
            name = as_group_name(self.resource_name, group)
            parts.append(f"```{header}\n{name}: {self.text}\n```")
        return CHANGELOG_SEPARATOR.join(parts)

    def __str__(self) -> str:
        return self.as_changelog()


def _convert_to_changelog(changes: str) -> str:
    if not changes.startswith("rs="):
        err = "missing rs=YOUR_RESOURCE_NAME e.g., mongodbatlas_federated_settings_org_config"
        raise ValueError(err)
    header, *change_lines = changes.splitlines()
    rs = header.removeprefix("rs=").strip()
    changelog = []
    for change in change_lines:
        if changelog_part := as_changelog_parts(rs, change):
            changelog.append(changelog_part)
    logger.info(f"found a total of {len(changelog)} changes")
    return CHANGELOG_SEPARATOR.join(str(part) for part in changelog)


def as_changelog_parts(resource_name: str, line_raw: str) -> ChangelogPart | None:
    if match := _line_regex.search(line_raw):
        return ChangelogPart(**match.groupdict(), resource_name=resource_name)  # type: ignore
    logger.warning(f"unable to parse line: {line_raw}")
    return None


def convert_to_changelog(changes: str) -> str:
    try:
        return _convert_to_changelog(changes)
    except ValueError as e:
        logger.critical(str(e))
        raise typer.Abort from e



import logging
from collections.abc import Iterable
from functools import singledispatch
from pathlib import Path
from typing import Annotated, Literal, NamedTuple

import pydantic
from model_lib import Entity, dump, field_names, parse_model
from zero_3rdparty import dict_nested
from zero_3rdparty.enum_utils import StrEnum

logger = logging.getLogger(__name__)


class ProviderSpecMapAttribute(Entity):
    computed_optional_required: Literal["computed_optional"]
    element_type: dict[str, dict]
    description: str


class ProviderSpecAttribute(Entity):
    name: str
    map: ProviderSpecMapAttribute | None = None

    def dump_provider_code_spec(self) -> dict:
        return self.model_dump(exclude_none=True)


class IgnoreNested(Entity):
    type: Literal["ignore_nested"] = "ignore_nested"
    path: str

    @property
    def use_wildcard(self) -> bool:
        return "*" in self.path


class RenameAttribute(Entity):
    type: Literal["rename_attribute"] = "rename_attribute"
    from_name: str
    to_name: str


class ComputedOptionalRequired(StrEnum):
    COMPUTED_OPTIONAL = "computed_optional"
    REQUIRED = "required"
    COMPUTED = "computed"
    OPTIONAL = "optional"


class ChangeAttributeType(Entity):
    type: Literal["change_attribute_type"] = "change_attribute_type"
    path: str
    new_value: ComputedOptionalRequired

    @classmethod
    def read_value(cls, attribute_dict: dict) -> str:
        return attribute_dict["string"]["computed_optional_required"]

    def update_value(self, attribute_dict: dict) -> None:
        attribute_dict["string"]["computed_optional_required"] = self.new_value


class SkipValidators(Entity):
    type: Literal["skip_validators"] = "skip_validators"


Extension = Annotated[
    IgnoreNested | RenameAttribute | ChangeAttributeType | SkipValidators,
    pydantic.Field("type"),
]


class TFResource(Entity):
    model_config = pydantic.ConfigDict(extra="allow")
    name: str
    extensions: list[Extension] = pydantic.Field(default_factory=list)
    provider_spec_attributes: list[ProviderSpecAttribute] = pydantic.Field(
        default_factory=list
    )

    def dump_generator_config(self) -> dict:
        names = field_names(self)
        return self.model_dump(exclude=set(names))


class PyTerraformSchema(Entity):
    resources: list[TFResource]
    data_sources: list[TFResource] = pydantic.Field(default_factory=list)

    def resource(self, resource: str) -> TFResource:
        return next(r for r in self.resources if r.name == resource)


def parse_py_terraform_schema(path: Path) -> PyTerraformSchema:
    return parse_model(path, PyTerraformSchema)


def dump_generator_config(schema: PyTerraformSchema) -> str:
    resources = {}
    for resource in schema.resources:
        resources[resource.name] = resource.dump_generator_config()
    data_sources = {ds.name: ds.dump_generator_config() for ds in schema.data_sources}
    generator_config = {
        "provider": {"name": "mongodbatlas"},
        "resources": resources,
        "data_sources": data_sources,
    }
    return dump(generator_config, "yaml")


class AttributeTuple(NamedTuple):
    name: str
    path: str
    attribute_dict: dict

    @property
    def attribute_path(self) -> str:
        return f"{self.path}.{self.name}" if self.path else self.name


class ProviderCodeSpec(Entity):
    model_config = pydantic.ConfigDict(extra="allow")
    provider: dict
    resources: list[dict]
    datasources: list[dict] = pydantic.Field(default_factory=list)
    version: str

    def root_dict(self, name: str, is_datasource: bool = False) -> dict:  # noqa: FBT002
        resources = self.datasources if is_datasource else self.resources
        root_value = next((r for r in resources if r["name"] == name), None)
        if root_value is None:
            raise ValueError(f"{self.root_name(name, is_datasource)} not found!")
        return root_value

    def schema_attributes(
        self, name: str, is_datasource: bool = False
    ) -> list:
        root_dict = self.root_dict(name, is_datasource)
        return root_dict["schema"]["attributes"]

    def _type_name(self, is_datasource: bool):
        return "datasource" if is_datasource else "resource"

    def root_name(self, name: str, is_datasource: bool):
        return f"{self._type_name(is_datasource)}.{name}"

    def attribute_names(
        self, name: str, is_datasource: bool = False
    ) -> list[str]:
        return [
            a["name"] for a in self.schema_attributes(name, is_datasource=is_datasource)
        ]

    def iter_all_attributes(
        self, name: str, is_datasource: bool = False
    ) -> Iterable[AttributeTuple]:
        for attribute in self.schema_attributes(name=name, is_datasource=is_datasource):
            yield AttributeTuple(attribute["name"], "", attribute)
        yield from self.iter_nested_attributes(name, is_datasource=is_datasource)

    def iter_nested_attributes(
        self, name: str, is_datasource: bool = False
    ) -> Iterable[AttributeTuple]:
        for i, attribute in enumerate(
            self.schema_attributes(name=name, is_datasource=is_datasource)
        ):
            for path, attr_dict in dict_nested.iter_nested_key_values(
                attribute, type_filter=dict, include_list_indexes=True
            ):
                full_path = f"[{i}].{path}"
                if name := attr_dict.get("name", ""):
                    yield AttributeTuple(name, full_path, attr_dict)

    def remove_nested_attribute(
        self, name: str, path: str, is_datasource: bool = False
    ) -> None:
        root_name = self.root_name(name, is_datasource)
        logger.info(f"will remove attribute from {root_name} with path: {path}")
        root_attributes = self.root_dict(name, is_datasource)
        full_path = f"schema.attributes.{path}"
        popped = dict_nested.pop_nested(root_attributes, full_path, "")
        if popped == "":
            raise ValueError(
                f"failed to remove attribute from resource {name} with path: {full_path}"
            )
        assert isinstance(
            popped, dict
        ), f"expected removed attribute to be a dict, got: {popped}"
        logger.info(f"removal ok, attribute_name: '{root_name}.{popped.get('name')}'")

    def read_attribute(
        self, name: str, path: str, *, is_datasource: bool = False
    ) -> dict:
        if "." not in path:
            attribute_dict = next(
                (
                    a
                    for a in self.schema_attributes(name, is_datasource)
                    if a["name"] == path
                ),
                None,
            )
        else:
            root_dict = self.root_dict(name, is_datasource)
            attribute_dict = dict_nested.read_nested_or_none(
                root_dict, f"schema.attributes.{path}"
            )
        if attribute_dict is None:
            raise ValueError(
                f"attribute {path} not found in {self.root_name(name, is_datasource)}"
            )
        assert isinstance(
            attribute_dict, dict
        ), f"expected attribute to be a dict, got: {attribute_dict} @ {path} for resource={name}"
        return attribute_dict


def update_provider_code_spec(
    schema: PyTerraformSchema, provider_code_spec_path: Path
) -> str:
    spec = parse_model(provider_code_spec_path, t=ProviderCodeSpec)
    for resource in schema.resources:
        resource_name = resource.name
        if extra_spec_attributes := resource.provider_spec_attributes:
            add_explicit_attributes(spec, resource_name, extra_spec_attributes)
        for extension in resource.extensions:
            apply_extension(extension, spec, resource_name)
    for data_source in schema.data_sources:
        data_source_name = data_source.name
        if extra_spec_attributes := data_source.provider_spec_attributes:
            add_explicit_attributes(
                spec, data_source_name, extra_spec_attributes, is_datasource=True
            )
        for extension in data_source.extensions:
            apply_extension(extension, spec, data_source_name, is_datasource=True)
    return dump(spec, "json")


def add_explicit_attributes(
    spec: ProviderCodeSpec,
    name: str,
    extra_spec_attributes: list[ProviderSpecAttribute],
    *,
    is_datasource=False,
):
    resource_attributes = spec.schema_attributes(name, is_datasource=is_datasource)
    existing_names = spec.attribute_names(name, is_datasource=is_datasource)
    new_names = [extra.name for extra in extra_spec_attributes]
    if both := set(existing_names) & set(new_names):
        raise ValueError(f"resource: {name}, has already: {both} attributes")
    resource_attributes.extend(
        extra.dump_provider_code_spec() for extra in extra_spec_attributes
    )


@singledispatch
def apply_extension(
    extension: object,
    spec: ProviderCodeSpec,
    resource_name: str,
    *,
    is_datasource: bool = False,
):
    raise NotImplementedError(f"unsupported extension: {extension!r}")


@apply_extension.register  # type: ignore
def _ignore_nested(
    extension: IgnoreNested,
    spec: ProviderCodeSpec,
    resource_name: str,
    *,
    is_datasource: bool = False,
):
    if extension.use_wildcard:
        name_to_remove = extension.path.removeprefix("*.")
        assert (
            "*" not in name_to_remove
        ), f"only prefix *. is allowed for wildcard in path {extension.path}"
        found_paths = [
            path
            for name, path, attribute_dict in spec.iter_nested_attributes(
                resource_name, is_datasource=is_datasource
            )
            if name == name_to_remove
        ]
        while found_paths:
            next_to_remove = found_paths.pop()
            spec.remove_nested_attribute(
                resource_name, next_to_remove, is_datasource=is_datasource
            )
            found_paths = [
                path
                for name, path, attribute_dict in spec.iter_nested_attributes(
                    resource_name, is_datasource=is_datasource
                )
                if name == name_to_remove
            ]
    else:
        err_msg = "only wildcard path is supported"
        raise NotImplementedError(err_msg)


@apply_extension.register  # type: ignore
def _rename_attribute(
    extension: RenameAttribute,
    spec: ProviderCodeSpec,
    resource_name: str,
    *,
    is_datasource: bool = False,
):
    for attribute_dict in spec.schema_attributes(
        resource_name, is_datasource=is_datasource
    ):
        if attribute_dict.get("name") == extension.from_name:
            logger.info(
                f"renaming attribute for {spec.root_name(resource_name, is_datasource)}: {extension.from_name} -> {extension.to_name}"
            )
            attribute_dict["name"] = extension.to_name


@apply_extension.register  # type: ignore
def _change_attribute_type(
    extension: ChangeAttributeType,
    spec: ProviderCodeSpec,
    resource_name: str,
    *,
    is_datasource: bool = False,
):
    attribute_dict = spec.read_attribute(
        resource_name, extension.path, is_datasource=is_datasource
    )
    old_value = extension.read_value(attribute_dict)
    if old_value == extension.new_value:
        logger.info(
            f"no change for '{spec.root_name(resource_name, is_datasource)}': {extension.path} -> {extension.new_value}"
        )
        return

    logger.info(
        f"changing attribute type for '{spec.root_name(resource_name, is_datasource)}.{extension.path}': {old_value} -> {extension.new_value}"
    )
    extension.update_value(attribute_dict)


@apply_extension.register  # type: ignore
def _skip_validators(
    _: SkipValidators,
    spec: ProviderCodeSpec,
    resource_name: str,
    *,
    is_datasource: bool = False,
):
    for attr_tuple in spec.iter_all_attributes(
        resource_name, is_datasource=is_datasource
    ):
        attribute_dict = attr_tuple.attribute_dict
        paths_to_pop = [
            f"{path}.validators"
            for path, nested_dict in dict_nested.iter_nested_key_values(
                attribute_dict, type_filter=dict
            )
            if "validators" in nested_dict
        ]
        if paths_to_pop:
            logger.info(f"popping validators from '{attr_tuple.attribute_path}'")
        for path in paths_to_pop:
            dict_nested.pop_nested(attribute_dict, path)



import re
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass


@dataclass(eq=True, frozen=False)
class Block:
    name: str
    line_start: int
    level: int
    hcl: str = ""
    line_end: int = -1

    def __post_init__(self):
        self.name = self.name.strip()

    def end_block(self, line_end: int, hcl: str):
        self.line_end = line_end
        self.hcl = hcl

    @property
    def _lines(self) -> list[str]:
        return self.hcl.splitlines()

    def content_lines(self) -> list[tuple[int, str]]:
        return list(enumerate(self._lines[1:-1], start=1))

    def read_lines(self, start: int, end: int) -> list[str]:
        return self._lines[start : end + 1]

    def __hash__(self) -> int:
        return hash((self.name, self.line_start, self.line_end))


@dataclass
class ResourceBlock(Block):
    type: str = ""

    def __post_init__(self):
        self.name = self.name.strip()
        self.type = self.type.strip()

    def __hash__(self) -> int:
        return hash((self.name, self.type))

    def __str__(self) -> str:
        return f"{self.resource_id} @ L{self.line_start}-{self.line_end}"

    @property
    def resource_id(self) -> str:
        return f"{self.type}.{self.name}"


_resource_pattern = re.compile(r"resource\s+\"(?P<type>[^\"]+)\"\s+\"(?P<name>[^\"]+)\"\s+\{")


def iter_resource_blocks(hcl_config: str) -> Iterable[ResourceBlock]:
    # support line_nr indexing
    lines = ["", *hcl_config.splitlines()]
    current_block = None
    for i, line in enumerate(lines):
        if current_block is not None:
            if line.rstrip() == "}":
                current_block.end_block(i, "\n".join(lines[current_block.line_start : i + 1]))
                yield current_block
                current_block = None
            continue
        if match := _resource_pattern.match(line):
            assert current_block is None, "Nested blocks resource blocks are not supported"
            current_block = ResourceBlock(
                name=match.group("name"),
                type=match.group("type"),
                line_start=i,
                level=0,
            )
    if current_block is not None:
        err_msg = "Final resource block not closed"
        raise ValueError(err_msg)


_block_pattern = re.compile(r"(?P<name>[^\{]+)[\s=]+\{")


def iter_blocks(block: Block, level: int | None = None) -> Iterable[Block]:
    level = level or block.level + 1
    line_level_start_names: dict[int, tuple[int, str]] = {}
    current_level = level
    for line_nr, line in block.content_lines():
        if match := _block_pattern.match(line):
            line_level_start_names[current_level] = (line_nr, match.group("name"))
            current_level += 1
        if line.strip() == "}":
            current_level -= 1
            start_line_nr_name = line_level_start_names.pop(current_level, None)
            if start_line_nr_name is None:
                raise ValueError(f"Unbalanced block @ {line_nr} in {block.name}")
            start_line_nr, name = start_line_nr_name
            if level == current_level:
                block_lines: list[str] = block.read_lines(start_line_nr, line_nr)
                if "=" in block_lines[0]:
                    continue
                yield Block(
                    name=name,
                    line_start=start_line_nr,
                    level=level,
                    line_end=line_nr,
                    hcl="\n".join(block_lines),
                )
    if line_level_start_names.get(level) is not None:
        raise ValueError(f"Unfinished block @ {line_nr} in {block.name} at level {level}")  # pyright: ignore


def hcl_attrs(block: Block) -> dict[str, str]:
    nested_blocks = list(iter_blocks(block, level=block.level + 1))
    block_lines = as_block_lines(nested_blocks)
    return _hcl_attrs(block, block_lines)


def _hcl_attrs(block: Block, block_lines: set[int]) -> dict[str, str]:
    attrs = defaultdict(list)
    attr_name: str | None = None
    for line_nr, line in block.content_lines():
        if line_nr in block_lines:
            continue
        if "=" in line:
            assert attr_name is None, f"unfinished attribute {attr_name}, new attribute at {line_nr}"
            attr_name, attr_value = line.split("=", 1)
            attrs[attr_name.strip()] = [attr_value.strip()]
            if line.rstrip().endswith(("{", "[", ",")):
                raise ValueError(f"unsupported nested attribute assignment on {line_nr} in block: {block.name}")
            attr_name = None
    return {k: "\n".join(v) for k, v in attrs.items()}


def as_block_lines(blocks: list[Block]) -> set[int]:
    block_lines = set()
    for block in blocks:
        block_lines.update(set(range(block.line_start, block.line_end)))
    return block_lines



from ask_shell import configure_logging
from typer import Typer

from atlas_init.tf_ext import api_call


def typer_main():
    from atlas_init.tf_ext import tf_dep, tf_modules, tf_vars

    app = Typer(
        name="tf-ext",
        help="Terraform extension commands for Atlas Init",
    )
    app.command(name="dep-graph")(tf_dep.tf_dep_graph)
    app.command(name="vars")(tf_vars.tf_vars)
    app.command(name="modules")(tf_modules.tf_modules)
    app.command(name="api")(api_call.api)
    app.command(name="api-config")(api_call.api_config)
    configure_logging(app)
    app()


if __name__ == "__main__":
    typer_main()



from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Sequence
from concurrent.futures import ThreadPoolExecutor, wait
from datetime import UTC, datetime
from functools import lru_cache, total_ordering
from pathlib import Path

import botocore.exceptions
import humanize
import typer
from boto3.session import Session
from model_lib import Event
from mypy_boto3_cloudformation import CloudFormationClient
from mypy_boto3_cloudformation.type_defs import ListTypesOutputTypeDef, ParameterTypeDef
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed
from zero_3rdparty.datetime_utils import utc_now
from zero_3rdparty.iter_utils import group_by_once

from atlas_init.cli_helper.run import run_command_is_ok
from atlas_init.cloud.aws import REGIONS, PascalAlias, region_continent
from atlas_init.settings.interactive import confirm

logger = logging.getLogger(__name__)
EARLY_DATETIME = datetime(year=1990, month=1, day=1, tzinfo=UTC)


@lru_cache
def cloud_formation_client(region_name: str = "") -> CloudFormationClient:
    return Session(region_name=region_name).client("cloudformation")  # type: ignore


def deregister_cfn_resource_type(type_name: str, deregister: bool, region_filter: str | None = None):
    for region in REGIONS:
        if region_filter and region != region_filter:
            continue
        try:
            default_version_arn = None
            client = cloud_formation_client(region)
            for version in client.list_type_versions(Type="RESOURCE", TypeName=type_name)["TypeVersionSummaries"]:
                logger.info(f"found version: {version} for {type_name} in {region}")
                if not deregister:
                    continue
                arn = version["Arn"]  # type: ignore
                if version["IsDefaultVersion"]:  # type: ignore
                    default_version_arn = arn.rsplit("/", maxsplit=1)[0]
                else:
                    logger.info(f"deregistering: {arn}")
                    client.deregister_type(Arn=arn)
            if default_version_arn is not None:
                logger.info(f"deregistering default-arn: {default_version_arn}")
                client.deregister_type(Arn=default_version_arn)
        except Exception as e:
            if "The type does not exist" in repr(e):
                logger.info(f"type={type_name} not found in {region}")
                continue
            raise


def deregister_arn(arn: str, region: str):
    client = cloud_formation_client(region)
    logger.warning(f"deregistering type {arn} in {region}")
    client.deregister_type(Arn=arn)


def deactivate_third_party_type(type_name: str, region_name: str, *, dry_run: bool = False) -> None | CfnTypeDetails:
    last_version = get_last_cfn_type(type_name, region=region_name, is_third_party=True)
    if not last_version:
        logger.info(f"no third party found in region {region_name}")
        return
    is_activated = last_version.is_activated
    logger.info(f"found {last_version.type_name} {last_version.version} in {region_name}, is_activated={is_activated}")
    if is_activated and not dry_run:
        deactivate_type(type_name=type_name, region=region_name)


def deactivate_type(type_name: str, region: str):
    client = cloud_formation_client(region)
    logger.warning(f"deactivating type {type_name} in {region}")
    client.deactivate_type(TypeName=type_name, Type="RESOURCE")


def delete_role_stack(type_name: str, region_name: str, role_arn: str = "") -> None:
    stack_name = type_name.replace("::", "-").lower() + "-role-stack"
    delete_stack(region_name, stack_name, role_arn)


def delete_stack(region_name: str, stack_name: str, role_arn: str = ""):
    client = cloud_formation_client(region_name)
    logger.warning(f"deleting stack {stack_name} in region={region_name}")
    try:
        client.update_termination_protection(EnableTerminationProtection=False, StackName=stack_name)
    except Exception as e:
        if "does not exist" in repr(e):
            logger.warning(f"stack {stack_name} not found")
            return
        raise
    client.delete_stack(StackName=stack_name, RoleARN=role_arn)
    wait_on_stack_ok(stack_name, region_name, expect_not_found=True)


def create_stack(
    stack_name: str,
    template_str: str,
    region_name: str,
    role_arn: str,
    parameters: Sequence[ParameterTypeDef],
    timeout_seconds: int = 300,
):
    client = cloud_formation_client(region_name)
    stack_id = client.create_stack(
        StackName=stack_name,
        TemplateBody=template_str,
        Parameters=parameters,
        RoleARN=role_arn,
    )
    logger.info(
        f"stack with name: {stack_name} created in {region_name} has id: {stack_id['StackId']} role_arn:{role_arn}"
    )
    wait_on_stack_ok(stack_name, region_name, timeout_seconds=timeout_seconds)


def update_stack(
    stack_name: str,
    template_str: str,
    region_name: str,
    role_arn: str,
    parameters: Sequence[ParameterTypeDef],
    timeout_seconds: int = 300,
):
    client = cloud_formation_client(region_name)
    update = client.update_stack(
        StackName=stack_name,
        TemplateBody=template_str,
        Parameters=parameters,
        RoleARN=role_arn,
    )
    logger.info(f"stack with name: {stack_name} updated {region_name} has id: {update['StackId']}")
    wait_on_stack_ok(stack_name, region_name, timeout_seconds=timeout_seconds)


class StackBaseError(Exception):
    def __init__(self, status: str, timestamp: datetime, status_reason: str) -> None:
        super().__init__(status, timestamp, status_reason)
        self.status = status
        self.timestamp = timestamp
        self.status_reason = status_reason


class StackInProgressError(StackBaseError):
    pass


class StackError(StackBaseError):
    def __init__(self, status: str, timestamp: datetime, status_reason: str, reasons: str) -> None:
        super().__init__(status, timestamp, status_reason)
        self.reasons = reasons


@total_ordering
class StackEvent(Event):
    model_config = PascalAlias
    logical_resource_id: str
    timestamp: datetime
    resource_status: str
    resource_status_reason: str = ""

    @property
    def in_progress(self) -> bool:
        return self.resource_status.endswith("IN_PROGRESS")

    @property
    def is_error(self) -> bool:
        return self.resource_status.endswith("FAILED")

    def __lt__(self, other) -> bool:
        if not isinstance(other, StackEvent):
            raise TypeError
        return self.timestamp < other.timestamp


class StackEvents(Event):
    model_config = PascalAlias
    stack_events: list[StackEvent]

    def current_stack_event(self, stack_name: str) -> StackEvent:
        sorted_events = sorted(self.stack_events)
        for event in reversed(sorted_events):
            if event.logical_resource_id == stack_name:
                return event
        raise ValueError(f"no events found for {stack_name}")

    def last_reason(self) -> str:
        for event in sorted(self.stack_events, reverse=True):
            if reason := event.resource_status_reason.strip():
                return reason
        return ""

    def multiple_reasons(self, max_reasons: int = 5) -> str:
        reasons = []
        for event in sorted(self.stack_events, reverse=True):
            if reason := event.resource_status_reason.strip():
                reason_number = len(reasons) + 1
                reasons.append(f"{reason_number}. {reason}")
                if reason_number >= max_reasons:
                    break
        return "\n".join(reasons)


def wait_on_stack_ok(
    stack_name: str,
    region_name: str,
    *,
    expect_not_found: bool = False,
    timeout_seconds: int = 300,
) -> None:
    attempts = timeout_seconds // 6

    @retry(
        stop=stop_after_attempt(attempts + 1),
        wait=wait_fixed(6),
        retry=retry_if_exception_type(StackInProgressError),
        reraise=True,
    )
    def _wait_on_stack_ok() -> None:
        client = cloud_formation_client(region_name)
        try:
            response = client.describe_stack_events(StackName=stack_name)
        except botocore.exceptions.ClientError as e:
            if not expect_not_found:
                raise
            error_message = e.response.get("Error", {}).get("Message", "")
            if "does not exist" not in error_message:
                raise
            return None
        parsed = StackEvents(stack_events=response.get("StackEvents", []))  # type: ignore
        current_event = parsed.current_stack_event(stack_name)
        if current_event.in_progress:
            logger.info(f"stack in progress {stack_name} {current_event.resource_status}")
            raise StackInProgressError(
                current_event.resource_status,
                current_event.timestamp,
                current_event.resource_status_reason,
            )
        if current_event.is_error:
            raise StackError(
                current_event.resource_status,
                current_event.timestamp,
                current_event.resource_status_reason,
                reasons=parsed.multiple_reasons(),
            )
        status = current_event.resource_status
        logger.info(f"stack is ready {stack_name} {status} ✅")
        if "ROLLBACK" in status:
            last_reason = parsed.multiple_reasons()
            logger.warning(f"stack did rollback, got: {current_event!r}\n{last_reason}")

        return None

    try:
        return _wait_on_stack_ok()
    except StackError as e:
        logger.warning(f"stack error {stack_name} {e.status} {e.status_reason}\n{e.reasons}")
        raise typer.Exit(1) from None


def print_version_regions(type_name: str) -> None:
    version_regions = get_last_version_all_regions(type_name)
    if regions_with_no_version := version_regions.pop(None, []):
        logger.warning(f"no version for {type_name} found in {regions_with_no_version}")
    for version in sorted(version_regions.keys()):  # type: ignore
        regions = sorted(version_regions[version])
        regions_comma_separated = ",".join(regions)
        logger.info(f"'{version}' is latest in {regions_comma_separated}\ncontinents:")
        for continent, cont_regions in group_by_once(regions, key=region_continent).items():
            continent_regions = ", ".join(sorted(cont_regions))
            logger.info(f"continent={continent}: {continent_regions}")


def get_last_version_all_regions(type_name: str) -> dict[str | None, list[str]]:
    futures = {}
    with ThreadPoolExecutor(max_workers=10) as pool:
        for region in REGIONS:
            future = pool.submit(get_last_cfn_type, type_name, region, is_third_party=True)
            futures[future] = region
        done, not_done = wait(futures.keys(), timeout=300)
        for f in not_done:
            logger.warning(f"timeout to find version in region = {futures[f]}")
    version_regions: dict[str | None, list[str]] = defaultdict(list)
    for f in done:
        region: str = futures[f]
        try:
            version = f.result()
        except Exception:
            logger.exception(f"failed to find version in region = {region}, error 👆")
            continue
        version_regions[version].append(region)
    return version_regions


@total_ordering
class CfnTypeDetails(Event):
    last_updated: datetime
    version: str
    type_name: str
    type_arn: str
    is_activated: bool

    def __lt__(self, other) -> bool:
        if not isinstance(other, CfnTypeDetails):
            raise TypeError
        return self.last_updated < other.last_updated

    def seconds_since_update(self) -> float:
        return (utc_now() - self.last_updated).total_seconds()


def publish_cfn_type(region: str):
    client: CloudFormationClient = cloud_formation_client(region)
    client.publish_type()


def get_last_cfn_type(
    type_name: str, region: str, *, is_third_party: bool = False, force_version: str = ""
) -> None | CfnTypeDetails:
    client: CloudFormationClient = cloud_formation_client(region)
    prefix = type_name
    logger.info(f"finding public 3rd party for '{prefix}' in {region}")
    visibility = "PUBLIC" if is_third_party else "PRIVATE"
    category = "THIRD_PARTY" if is_third_party else "REGISTERED"
    type_details: list[CfnTypeDetails] = []
    kwargs = {
        "Visibility": visibility,
        "Filters": {"Category": category, "TypeNamePrefix": prefix},
        "MaxResults": 100,
    }
    next_token = ""  # nosec
    for _ in range(100):
        types_response: ListTypesOutputTypeDef = client.list_types(**kwargs)  # type: ignore
        next_token = types_response.get("NextToken", "")
        kwargs["NextToken"] = next_token
        for t in types_response["TypeSummaries"]:
            last_updated = t.get("LastUpdated", EARLY_DATETIME)
            last_version = t.get("LatestPublicVersion", "unknown-version")
            arn = t.get("TypeArn", "unknown_arn")
            detail = CfnTypeDetails(
                last_updated=last_updated,
                version=last_version,
                type_name=t.get("TypeName", type_name),
                type_arn=arn,
                is_activated=t.get("IsActivated", False),
            )
            if detail.type_name != type_name:
                continue
            type_details.append(detail)
            logger.debug(f"{last_version} published @ {last_updated}")
        if not next_token:
            break
    if not type_details:
        logger.warning(f"no version for {type_name} in region {region}")
        return None
    if force_version:
        for detail in type_details:
            if detail.version == force_version:
                return detail
        versions = [d.version for d in type_details]
        raise ValueError(f"unable to find version {force_version} for {type_name}, got {versions}")
    return sorted(type_details)[-1]


def activate_resource_type(details: CfnTypeDetails, region: str, execution_role_arn: str):
    client = cloud_formation_client(region)
    response = client.activate_type(
        Type="RESOURCE",
        PublicTypeArn=details.type_arn,
        ExecutionRoleArn=execution_role_arn,
    )
    logger.info(f"activate response: {response} role={execution_role_arn}")


def ensure_resource_type_activated(
    type_name: str,
    region: str,
    force_deregister: bool,
    is_interactive: bool,
    resource_path: Path,
    cfn_execution_role: str,
    force_version: str = "",
) -> None:
    cfn_type_details = get_last_cfn_type(type_name, region, is_third_party=False)
    logger.info(f"found cfn_type_details {cfn_type_details} for {type_name}")
    is_third_party = False
    if cfn_type_details is None:
        cfn_type_details = get_last_cfn_type(type_name, region, is_third_party=True)
        if cfn_type_details:
            is_third_party = True
            logger.warning(f"found 3rd party extension for cfn type {type_name} active")
    if force_version:
        if cfn_type_details and cfn_type_details.version == force_version:
            logger.info(f"version {force_version} already active")
            return
        force_deregister = True
    if cfn_type_details is not None and (cfn_type_details.seconds_since_update() > 3600 * 24 or force_deregister):
        outdated_warning = f"more than {humanize.naturaldelta(cfn_type_details.seconds_since_update())} since last update to {type_name} {cfn_type_details.version}"
        logger.warning(outdated_warning)
        if force_deregister or confirm(
            f"{outdated_warning}, should deregister?",
            is_interactive=is_interactive,
            default=True,
        ):
            if is_third_party:
                deactivate_third_party_type(type_name, region)
            else:
                deregister_cfn_resource_type(type_name, deregister=True, region_filter=region)
            cfn_type_details = None

    # assert cfn_type_details, f"no cfn_type_details found for {type_name}"
    # client = cloud_formation_client(region)
    # response = client.activate_type(
    #     Type="RESOURCE",
    #     # PublicTypeArn=cfn_type_details.type_arn.replace(":type/", "::type/"),
    #     PublicTypeArn="arn:aws:cloudformation:eu-south-2:358363220050::type/resource/MongoDB-Atlas-ResourcePolicy/00000001",
    #     ExecutionRoleArn=cfn_execution_role,
    #     LoggingConfig={"LogRoleArn": cfn_execution_role, "LogGroupName": "/apix/espen/cfn-test1"},
    # )
    # logger.info(f"activate response: {response}")
    submit_cmd = f"cfn submit --verbose --set-default --region {region} --role-arn {cfn_execution_role}"
    if (
        not force_version
        and cfn_type_details is None
        and confirm(
            f"No existing {type_name} found, ok to run:\n{submit_cmd}\nsubmit?",
            is_interactive=is_interactive,
            default=True,
        )
    ):
        assert run_command_is_ok(cmd=submit_cmd, env=None, cwd=resource_path, logger=logger)
        cfn_type_details = get_last_cfn_type(type_name, region, is_third_party=False)
    if cfn_type_details is None:
        third_party = get_last_cfn_type(type_name, region, is_third_party=True, force_version=force_version)
        assert third_party, f"unable to find 3rd party type for {type_name}"
        last_updated = third_party.last_updated
        if confirm(
            f"No existing {type_name} found, ok to activate 3rd party: :\n'{third_party.version} ({humanize.naturalday(last_updated), {last_updated.isoformat()}})'\n?",
            is_interactive=is_interactive,
            default=True,
        ):
            activate_resource_type(third_party, region, cfn_execution_role)
            cfn_type_details = third_party
    assert cfn_type_details, f"no cfn_type_details found for {type_name}"
    # TODO: validate the active type details uses the execution role



from pydantic import BaseModel, Field
from typing import Dict, List, Optional
from enum import Enum


class HttpMethod(str, Enum):
    """HTTP methods enum"""

    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    DELETE = "DELETE"
    PUT = "PUT"


class WaitConfig(BaseModel):
    """Configuration for waiting/polling operations"""

    state_property: str = Field(..., description="Property to check for state")
    pending_states: List[str] = Field(..., description="States that indicate operation is still in progress")
    target_states: List[str] = Field(..., description="States that indicate operation is complete")
    timeout_seconds: int = Field(..., description="Maximum time to wait in seconds")
    min_timeout_seconds: int = Field(..., description="Minimum timeout in seconds")
    delay_seconds: int = Field(..., description="Delay between polling attempts in seconds")


class OperationConfig(BaseModel):
    """Configuration for a single API operation (read, create, update, delete)"""

    path: str = Field(..., description="API endpoint path with path parameters")
    method: HttpMethod = Field(..., description="HTTP method for the operation")
    wait: Optional[WaitConfig] = Field(None, description="Wait configuration for async operations")


class SchemaOverrides(BaseModel):
    """Schema overrides for specific fields"""

    sensitive: Optional[bool] = Field(None, description="Mark field as sensitive")
    # Add other override properties as needed


class SchemaConfig(BaseModel):
    """Schema configuration for the resource"""

    aliases: Optional[Dict[str, str]] = Field(None, description="Field name aliases mapping")
    overrides: Optional[Dict[str, SchemaOverrides]] = Field(None, description="Field-specific overrides")
    ignores: Optional[List[str]] = Field(None, description="Fields to ignore")
    timeouts: Optional[List[str]] = Field(None, description="Operations that support timeouts")


class ResourceConfig(BaseModel):
    """Configuration for a single API resource"""

    read: Optional[OperationConfig] = Field(None, description="Read operation configuration")
    create: Optional[OperationConfig] = Field(None, description="Create operation configuration")
    update: Optional[OperationConfig] = Field(None, description="Update operation configuration")
    delete: Optional[OperationConfig] = Field(None, description="Delete operation configuration")
    version_header: Optional[str] = Field(None, description="API version header value")
    custom_schema: Optional[SchemaConfig] = Field(None, description="Schema configuration", alias="schema")

    @property
    def paths(self) -> list[str]:
        return [
            operation.path
            for operation in [self.read, self.create, self.update, self.delete]
            if operation and operation.path
        ]


class ApiResourcesConfig(BaseModel):
    """Root configuration model containing all API resources"""

    resources: Dict[str, ResourceConfig] = Field(..., description="Dictionary of resource configurations")

    class Config:
        extra = "allow"  # Allow additional fields not explicitly defined

    def get_resource(self, name: str) -> ResourceConfig:
        """Get a specific resource configuration by name"""
        resource = self.resources.get(name)
        if not resource:
            raise ValueError(f"Resource '{name}' not found in configuration")
        return resource

    def list_resources(self) -> List[str]:
        """Get list of all resource names"""
        return list(self.resources.keys())

    def get_resources_with_operation(self, operation: str) -> List[str]:
        """Get list of resources that support a specific operation"""
        result = []
        result.extend(
            name
            for name, config in self.resources.items()
            if hasattr(config, operation) and getattr(config, operation) is not None
        )
        return result



import logging
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor, wait
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Annotated, TypeVar

import stringcase
from pydantic import AfterValidator, ConfigDict
from zero_3rdparty.file_utils import copy, file_modified_time, iter_paths_and_relative
from zero_3rdparty.iter_utils import flat_map
from zero_3rdparty.object_name import as_name

from atlas_init.cli_helper.run import run_binary_command_is_ok
from atlas_init.cli_root import is_dry_run

logger = logging.getLogger(__name__)
PascalAlias = ConfigDict(alias_generator=stringcase.pascalcase, populate_by_name=True)
REGIONS = "af-south-1,ap-east-1,ap-northeast-1,ap-northeast-2,ap-northeast-3,ap-south-1,ap-southeast-1,ap-southeast-2,ap-southeast-3,ca-central-1,eu-central-1,eu-north-1,eu-south-1,eu-west-1,eu-west-2,eu-west-3,me-south-1,sa-east-1,us-east-1,us-east-2,us-west-1,us-west-2,ap-south-2,ap-southeast-4,eu-central-2,eu-south-2,me-central-1,il-central-1".split(
    ","
)
REGION_CONTINENT_PREFIXES = {
    "Americas": ["us", "ca", "sa"],
    "Asia Pacific": ["ap"],
    "Europe": ["eu"],
    "Middle East and Africa": ["me", "af", "il"],
}
REGION_PREFIX_CONTINENT = dict(
    flat_map(
        [[(prefix, continent) for prefix in prefixes] for continent, prefixes in REGION_CONTINENT_PREFIXES.items()]
    )
)


def region_continent(region: str) -> str:
    """based on: https://www.mongodb.com/docs/atlas/reference/amazon-aws/"""
    prefix = region.split("-", maxsplit=1)[0]
    return REGION_PREFIX_CONTINENT.get(prefix, "UNKNOWN_CONTINENT")


def check_region_found(region: str) -> str:
    if region not in REGIONS:
        raise ValueError(f"unknown region: {region}")
    return region


AwsRegion = Annotated[str, AfterValidator(check_region_found)]
T = TypeVar("T")


def run_in_regions(call: Callable[[str], T], regions: list[str] | None = None) -> dict[str, T]:  # type: ignore
    futures = {}
    name = as_name(call)
    regions: list[str] = regions or REGIONS  # type: ignore
    with ThreadPoolExecutor(max_workers=10) as pool:
        for region in regions:
            future = pool.submit(call, region)
            futures[future] = region
        done, not_done = wait(futures.keys(), timeout=300)
        for f in not_done:
            logger.warning(f"timeout for {name} in region = {futures[f]}")
    region_responses: dict[str, T] = {}
    for f in done:
        region: str = futures[f]
        try:
            response = f.result()
            region_responses[region] = response
        except Exception:
            logger.exception(f"failed to call {name} in region = {region}, error 👆")
    return region_responses


def upload_to_s3(profile_path: Path, s3_bucket: str, s3_prefix: str = ""):
    profiles_path = profile_path.parent
    assert profiles_path.name == "profiles"
    excluded = [".DS_Store", ".terraform/*"]
    excluded_str = " ".join([f'--exclude "{pattern}"' for pattern in excluded])
    dest_path = _s3_path(s3_bucket, profile_path.name, "", s3_prefix=s3_prefix)
    assert run_binary_command_is_ok(
        "aws",
        f"s3 sync {profile_path.name} {dest_path} {excluded_str}",
        profiles_path,
        logger=logger,
        dry_run=is_dry_run(),
    )


def _s3_path(s3_bucket: str, profile_name: str, rel_path: str, s3_prefix: str = "") -> str:
    return f"s3://{s3_bucket}//{s3_prefix}profiles/{profile_name}/{rel_path}"


def download_from_s3(profile_path: Path, s3_bucket: str, s3_prefix: str = ""):
    profiles_path = profile_path.parent
    assert profiles_path.name == "profiles"
    src_path = _s3_path(s3_bucket, profile_path.name, "", s3_prefix)
    copy_dir = f"{profile_path.name}_copy/"
    with TemporaryDirectory() as tmp_dir:
        copy_dir = Path(tmp_dir) / f"safe-{profile_path.name}"
        assert run_binary_command_is_ok(
            "aws",
            f"s3 sync {src_path} {copy_dir}",
            profiles_path,
            logger=logger,
            dry_run=is_dry_run(),
        )
        copy_new_files(copy_dir, profile_path)


_aws_keys = (
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_PROFILE",
)


def copy_new_files(src_dir: Path, dest_dir: Path):
    for src_path, rel_path in iter_paths_and_relative(src_dir, "*", only_files=True):
        dest_path = dest_dir / rel_path
        if not dest_path.exists() or file_modified_time(src_path) > file_modified_time(dest_path):
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            if src_path.name == ".env-manual":
                if dest_path.exists():
                    continue  # never overwrite the manual file
                lines_no_aws = [line for line in src_path.read_text().splitlines() if not line.startswith(_aws_keys)]
                dest_path.write_text("\n".join(lines_no_aws) + "\n")
            else:
                copy(src_path, dest_path)



import logging
from pathlib import Path
from typing import Any

import typer
from pydantic import field_validator, model_validator
from rich import prompt

from atlas_init.cli_args import parse_key_values_any
from atlas_init.cli_cfn.aws import (
    create_stack,
    ensure_resource_type_activated,
    update_stack,
)
from atlas_init.cli_cfn.aws import delete_stack as delete_stack_aws
from atlas_init.cli_cfn.cfn_parameter_finder import (
    CfnTemplate,
    dump_resource_to_file,
    dump_sample_file,
    infer_template_parameters,
    infer_template_path,
)
from atlas_init.repos.cfn import CfnType, Operation, infer_cfn_type_name
from atlas_init.repos.path import Repo, find_paths
from atlas_init.settings.env_vars import AtlasInitSettings, init_settings
from atlas_init.settings.env_vars_generated import AWSSettings
from atlas_init.settings.env_vars_modules import TFModuleCfn

logger = logging.getLogger(__name__)


class CfnExampleInputs(CfnType):
    stack_name: str
    operation: Operation
    example_name: str
    resource_params: dict[str, Any] | None = None
    stack_timeout_s: int
    delete_stack_first: bool
    reg_version: str = ""
    force_deregister: bool
    force_keep: bool
    execution_role: str
    export_example_to_inputs: bool
    export_example_to_samples: bool
    register_all_types_in_example: bool

    @field_validator("resource_params", mode="before")
    @classmethod
    def validate_resource_params(cls, v):
        return parse_key_values_any(v) if isinstance(v, list) else v

    @model_validator(mode="after")
    def check(self):
        assert self.region_filter, "region is required"
        assert self.execution_role.startswith("arn:aws:iam::"), f"invalid execution role: {self.execution_role}"
        assert self.region
        if self.delete_stack_first and self.operation == Operation.UPDATE:
            err_msg = "cannot delete first when updating"
            raise ValueError(err_msg)
        return self

    @property
    def is_export(self) -> bool:
        return self.export_example_to_inputs or self.export_example_to_samples

    @property
    def region(self) -> str:
        region = self.region_filter
        assert isinstance(region, str), "region is required"
        return region


def example_cmd(
    type_name: str = typer.Option("", "-n", "--type-name", help="inferred from your cwd if not provided"),
    region: str = typer.Option("", help="inferred from your atlas_init cfn region if not provided"),
    stack_name: str = typer.Option(
        "",
        help="inferred from your atlas_init cfn profile name and example if not provided",
    ),
    operation: str = typer.Argument(...),
    example_name: str = typer.Option("", "-e", "--example-name", help="example filestem"),
    resource_params: list[str] = typer.Option(
        ...,
        "-r",
        "--resource-param",
        default_factory=list,
        help="key=value, can be set many times",
    ),
    stack_timeout_s: int = typer.Option(3600, "-t", "--stack-timeout-s"),
    delete_first: bool = typer.Option(False, "-d", "--delete-first", help="Delete existing stack first"),
    reg_version: str = typer.Option("", "--reg-version", help="Register a specific version"),
    force_deregister: bool = typer.Option(False, "--dereg", help="Force deregister CFN Type"),
    force_keep: bool = typer.Option(False, "--noreg", help="Force keep CFN Type (do not prompt)"),
    execution_role: str = typer.Option("", "--execution-role", help="Execution role to use, otherwise inferred"),
    export_example_to_inputs: bool = typer.Option(
        False, "-o", "--export-example-to-inputs", help="Export example to inputs"
    ),
    export_example_to_samples: bool = typer.Option(
        False, "-s", "--export-example-to-samples", help="Export example to samples"
    ),
    register_all_types_in_example: bool = typer.Option(False, "--reg-all", help="Check all types"),
):
    settings = init_settings(TFModuleCfn, AWSSettings)
    cfn_settings = TFModuleCfn.from_env()
    aws_settings = AWSSettings.from_env()
    assert settings.tf_vars, "no cfn config found, re-run atlas_init apply with CFN flags"
    repo_path, resource_path, _ = find_paths(Repo.CFN)
    inputs = CfnExampleInputs(
        type_name=type_name or infer_cfn_type_name(),
        example_name=example_name,
        delete_stack_first=delete_first,
        region_filter=region or settings.cfn_region(aws_settings.AWS_REGION),
        stack_name=stack_name or f"{cfn_settings.MONGODB_ATLAS_PROFILE}-{example_name or 'atlas-init'}",
        operation=operation,  # type: ignore
        resource_params=resource_params,  # type: ignore
        stack_timeout_s=stack_timeout_s,
        force_deregister=force_deregister,
        reg_version=reg_version,
        force_keep=force_keep,
        execution_role=execution_role or cfn_settings.CFN_EXAMPLE_EXECUTION_ROLE,
        export_example_to_inputs=export_example_to_inputs,
        export_example_to_samples=export_example_to_samples,
        register_all_types_in_example=register_all_types_in_example,
    )
    example_handler(inputs, repo_path, resource_path, settings)


def example_handler(
    inputs: CfnExampleInputs,
    repo_path: Path,
    resource_path: Path,
    settings: AtlasInitSettings,
):
    logger.info(
        f"about to {inputs.operation} stack {inputs.stack_name} for {inputs.type_name} in {inputs.region_filter} params: {inputs.resource_params}"
    )
    type_name = inputs.type_name
    stack_name = inputs.stack_name
    region = inputs.region
    operation = inputs.operation
    stack_timeout_s = inputs.stack_timeout_s
    delete_first = inputs.delete_stack_first
    force_deregister = inputs.force_deregister
    execution_role = inputs.execution_role

    template_path = infer_template_path(repo_path, type_name, stack_name, inputs.example_name)
    parameters = infer_template_parameters(template_path, type_name, stack_name, inputs.resource_params or {})
    logger.info(f"parameters: {parameters}")
    if not prompt.Confirm("parameters 👆looks good?")():
        raise typer.Abort

    logger.info(f"using execution role: {execution_role}")
    if not inputs.is_export and not inputs.force_keep:
        ensure_resource_type_activated(
            type_name,
            region,
            force_deregister,
            settings.is_interactive,
            resource_path,
            execution_role,
            force_version=inputs.reg_version,
        )
    if not inputs.is_export and (operation == Operation.DELETE or delete_first):
        delete_stack_aws(region, stack_name, execution_role)
        if not delete_first:
            return
    if inputs.register_all_types_in_example:
        extra_example_types = [t for t in CfnTemplate.read_template_types(template_path) if t != type_name]
        for extra_type in extra_example_types:
            logger.info(f"extra type {extra_type} in example {template_path}")
            ensure_resource_type_activated(
                extra_type,
                region,
                force_deregister,
                settings.is_interactive,
                resource_path,
                execution_role,
            )
    if inputs.export_example_to_inputs:
        out_inputs = dump_resource_to_file(resource_path / "inputs", template_path, type_name, parameters)
        logger.info(f"dumped to {out_inputs} ✅")
        return
    if inputs.export_example_to_samples:
        samples_dir = resource_path / "samples"
        samples_path = dump_sample_file(samples_dir, template_path, type_name, parameters)
        logger.info(f"dumped to {samples_path} ✅")
        return
    if operation == Operation.CREATE:
        create_stack(
            stack_name,
            template_str=template_path.read_text(),
            region_name=region,
            role_arn=execution_role,
            parameters=parameters,
            timeout_seconds=stack_timeout_s,
        )
    elif operation == Operation.UPDATE:
        update_stack(
            stack_name,
            template_str=template_path.read_text(),
            region_name=region,
            parameters=parameters,
            role_arn=execution_role,
            timeout_seconds=stack_timeout_s,
        )
    else:
        raise NotImplementedError



import random
from typing import Self

from pydantic import ConfigDict, Field, model_validator
from pydantic_settings import BaseSettings


class _EnvVarsGenerated(BaseSettings):
    model_config = ConfigDict(extra="ignore")  # type: ignore

    @classmethod
    def from_env(cls) -> Self:
        return cls()


class AtlasSettings(_EnvVarsGenerated):
    MONGODB_ATLAS_ORG_ID: str
    MONGODB_ATLAS_PRIVATE_KEY: str
    MONGODB_ATLAS_PUBLIC_KEY: str
    MONGODB_ATLAS_BASE_URL: str

    @property
    def realm_url(self) -> str:
        assert not self.is_mongodbgov_cloud, "realm_url is not supported for mongodbgov cloud"
        if "cloud-dev." in self.MONGODB_ATLAS_BASE_URL:
            return "https://services.cloud-dev.mongodb.com/"
        return "https://services.cloud.mongodb.com/"

    @property
    def is_mongodbgov_cloud(self) -> bool:
        return "mongodbgov" in self.MONGODB_ATLAS_BASE_URL


class AWSSettings(_EnvVarsGenerated):
    AWS_REGION: str
    AWS_PROFILE: str = ""
    AWS_ACCESS_KEY_ID: str = ""
    AWS_SECRET_ACCESS_KEY: str = ""

    @model_validator(mode="after")
    def ensure_credentials_are_given(self) -> Self:
        assert self.AWS_PROFILE or (self.AWS_ACCESS_KEY_ID and self.AWS_SECRET_ACCESS_KEY), (
            "Either AWS_PROFILE or both AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY must be provided"
        )
        assert not (self.AWS_PROFILE and (self.AWS_ACCESS_KEY_ID and self.AWS_SECRET_ACCESS_KEY)), (
            "Either AWS_PROFILE or both AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY must be provided, not both"
        )
        return self


class TerraformSettings(_EnvVarsGenerated):
    TF_CLI_CONFIG_FILE: str


class RealmSettings(_EnvVarsGenerated):
    MONGODB_REALM_APP_ID: str
    MONGODB_REALM_SERVICE_ID: str
    MONGODB_REALM_FUNCTION_ID: str
    MONGODB_REALM_FUNCTION_NAME: str
    MONGODB_REALM_BASE_URL: str
    RANDOM_INT_100K: str = Field(default_factory=lambda: str(random.randint(0, 100_000)))  # noqa: S311 # not used for cryptographic purposes # nosec


class AtlasSettingsWithProject(AtlasSettings):
    MONGODB_ATLAS_PROJECT_ID: str
    MONGODB_ATLAS_PROJECT_OWNER_ID: str = ""
    MONGODB_ATLAS_USER_EMAIL: str = ""



from functools import singledispatch
from typing import NamedTuple

from atlas_init.cli_tf.schema_v2_sdk import SDKAttribute
from atlas_init.cli_tf.schema_v3 import Attribute, Resource
from atlas_init.humps import pascalize


class SDKAndSchemaAttribute(NamedTuple):
    sdk_attribute: SDKAttribute
    schema_attribute: Attribute


_allowed_missing_attributes: set[str] = set()


def set_allowed_missing_attributes(allowed: set[str]):
    global _allowed_missing_attributes  # noqa: PLW0603 `Using the global statement to update `_allowed_missing_attributes` is discouraged`
    _allowed_missing_attributes = allowed


class AllowedMissingAttributeError(Exception):
    def __init__(self, name: str, *args: object) -> None:
        self.name = name
        super().__init__(*args)


def find_attribute(attributes: list[Attribute], name: str, root_name: str) -> Attribute:
    for schema_attribute in attributes:
        if name == schema_attribute.name:
            return schema_attribute
    if name in _allowed_missing_attributes:
        raise AllowedMissingAttributeError(name)
    raise ValueError(f"could not find schema attribute for {name} on resource: {root_name}")


@singledispatch
def schema_attributes(root: object) -> list[Attribute]:
    raise NotImplementedError(f"unsupported root type: {type(root)} to find schema attributes")


@schema_attributes.register
def _resource_attributes(root: Resource) -> list[Attribute]:
    return root.local_schema.attributes


@schema_attributes.register
def _attribute_nested(root: Attribute) -> list[Attribute]:
    return root.nested_attributes


def name_schema_struct(name: str) -> str:
    return f"TF{pascalize(name)}Model"


def name_struct_attribute(name: str) -> str:
    default = pascalize(name)
    if override := _name_attribute_overrides.get(default):
        return override
    return default


_name_attribute_overrides = {}


def set_name_attribute_overrides(overrides: dict[str, str]):
    global _name_attribute_overrides  # noqa: PLW0603 `Using the global statement to update `_name_attribute_overrides` is discouraged`
    _name_attribute_overrides = overrides



from __future__ import annotations

import json
import logging
import os
import subprocess
import sys
from collections.abc import Iterable
from pathlib import Path
from tempfile import TemporaryDirectory

from atlas_init.cli_tf.hcl.cluster_mig import (
    LEGACY_CLUSTER_TYPE,
    NEW_CLUSTER_TYPE,
    convert_clusters,
)
from atlas_init.cli_tf.hcl.parser import (
    ResourceBlock,
)

logger = logging.getLogger(__name__)


def should_continue(is_interactive: bool, question: str):
    for h in logger.handlers:
        h.flush()
    return input(f"{question} [y/N]") == "y" if is_interactive else True


def tf(cmd: str, tf_dir: Path, err_msg: str, err_msg_code_2: str = "") -> str:
    err_codes = {
        1: err_msg,
        2: err_msg_code_2,
    }
    with TemporaryDirectory() as temp_dir:
        result_file = Path(temp_dir) / "file"
        with open(result_file, "w") as file:
            exit_code = subprocess.call(
                f"terraform {cmd}".split(),
                stdin=sys.stdin,
                stderr=sys.stderr,
                stdout=file,
                cwd=tf_dir,
            )
            cmd_output = result_file.read_text().strip()
            if exit_code == 0:
                return cmd_output
            err_msg = err_codes.get(exit_code, err_msg) or err_msg
            logger.error(cmd_output)
            logger.error(err_msg)
            sys.exit(exit_code)


def convert_and_validate(tf_dir: Path, *, is_interactive: bool = False):
    out_path = tf_dir / "conversion_cluster_adv_cluster.tf"
    if out_path.exists() and should_continue(is_interactive, f"{out_path} already exists, should it be overwritten?"):
        logger.info(f"removing existing conversion file @ {out_path}")
        out_path.unlink()
    if adv_clusters := read_import_id_addresses(tf_dir, NEW_CLUSTER_TYPE):
        existing_addresses = ", ".join(adv_clusters.values())
        if should_continue(
            is_interactive,
            f"found existing advanced clusters: {existing_addresses}, should they be removed?",
        ):
            remove_from_state(tf_dir, adv_clusters.values())
    ensure_no_plan_changes(tf_dir)
    new_clusters_path = convert_clusters(tf_dir, out_path)
    logger.info(
        f"found a total of {len(new_clusters_path)} clusters to convert and generated their config to {out_path}"
    )
    if should_continue(is_interactive, f"should import the new clusters in {out_path}?"):
        import_new_clusters(tf_dir)
        ensure_no_plan_changes(tf_dir)
    else:
        logger.info("skipping import")
    if should_continue(is_interactive, "should replace the old cluster resources with the new ones?"):
        replace_old_clusters(tf_dir, out_path, new_clusters_path)
        ensure_no_plan_changes(tf_dir)
        logger.info(f"migration successful, migrated {len(new_clusters_path)} clusters!")
    else:
        logger.info("skipping replacment")


def remove_from_state(tf_dir, addresses: Iterable[str]) -> None:
    for address in addresses:
        logger.info(f"removing {address} from state")
        tf(f"state rm {address}", tf_dir, f"failed to remove {address}")


def ensure_no_plan_changes(tf_dir):
    logger.info("running plan to ensure there are no changes")
    tf(
        "plan -detailed-exitcode",
        tf_dir,
        "error running terraform plan",
        "plan had changes",
    )


def import_new_clusters(tf_dir: Path) -> None:
    cluster_import_ids = read_import_id_addresses(tf_dir)
    for import_id, resource_address in cluster_import_ids.items():
        new_resource_address = resource_address.replace(LEGACY_CLUSTER_TYPE, NEW_CLUSTER_TYPE)
        logger.info(f"importing {import_id} to {new_resource_address}")
        tf(
            f"import {new_resource_address} {import_id}",
            tf_dir,
            f"failed to import {new_resource_address}",
        )


def read_import_id_addresses(tf_dir: Path, resource_type: str = "") -> dict[str, str]:
    current_state = tf("show -json", tf_dir, "failed to read terraform state")
    return read_cluster_import_ids(current_state, resource_type)


def replace_old_clusters(
    tf_dir: Path,
    out_path: Path,
    new_clusters_path: dict[tuple[Path, ResourceBlock], str],
) -> None:
    out_path.unlink()
    for (path, block), new_config in new_clusters_path.items():
        old_resource_id = block.resource_id
        logger.info(f"replacing {old_resource_id} @ {path}")
        old_text = path.read_text()
        new_text = old_text.replace(block.hcl, new_config)
        path.write_text(new_text)
    remove_from_state(tf_dir, read_import_id_addresses(tf_dir).values())


def read_cluster_import_ids(state: str, resource_type: str = "") -> dict[str, str]:
    resource_type = resource_type or LEGACY_CLUSTER_TYPE
    try:
        json_state = json.loads(state)
    except json.JSONDecodeError:
        logger.exception("unable to decode state")
        sys.exit(1)
    resources = json_state["values"]["root_module"]["resources"]
    assert isinstance(resources, list)
    import_ids = {}
    for resource in resources:
        if resource["type"] == resource_type:
            project_id = resource["values"]["project_id"]
            name = resource["values"]["name"]
            import_id = f"{project_id}-{name}"
            import_ids[import_id] = resource["address"]
    return import_ids


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    *_, tf_dir_str = sys.argv
    tf_dir_path = Path(tf_dir_str)
    assert tf_dir_path.is_dir(), f"not a directory: {tf_dir_path}"
    fast_forward = os.environ.get("FAST_FORWARD", "false").lower() in {
        "yes",
        "true",
        "1",
    }
    convert_and_validate(tf_dir_path, is_interactive=not fast_forward)
